"""

:platform: Unix, Windows 
:last changed: 2024-01-24

.. moduleauthor:: Tjark Leon Raphael Gröne <tgroene@physnet.uni-hamburg.de>


"""

import pathlib
rootdir = str(pathlib.Path(__file__).parent.resolve())

from scipy import io as sp_io

from backbone.functions.plotting import Plottin_PDF
from backbone.functions.export import Export_PDF_Plot
from backbone.functions.import_data import Import_PDF_Plot
from backbone.functions.utils import averageArray


#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
#S/////////////////////////////////////////////////////////////////////////////////////////////////////S#
#S//                                                                                                 //S#
#S//    |----------------------------------------------------------------------------------------|   //S#
#S//    |                                                                                        |   //S#
#S//    |                                 The Insitu PDF Class                                   |   //S#
#S//    |                                                                                        |   //S#
#S//    |----------------------------------------------------------------------------------------|   //S#
#S//                                                                                                 //S#
#S/////////////////////////////////////////////////////////////////////////////////////////////////////S#
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#


class InsituPDF(Plottin_PDF,Export_PDF_Plot,Import_PDF_Plot):
    """ Class for plotting and exporting of the generated PDFs.

    Args:
        Plottin_PDF (class): Contains all the plotting functions.
        Export_PDF_Plot (class): Contains all the exporting functions.
        Import_PDF_Plot (class): Contains all the importing functions.
    """
    
    #----------------------------------------------------------------------------------------#
    #                                           Init                                         #
    #----------------------------------------------------------------------------------------#
    def __init__(self, path: str,tempPath: str,sampleName: str,startScan: int,endScan: int, timeOfset=0, 
                figratio1D=[16,9], figratio2D=[7, 4], reload: str = None, average: int = 30, without_dark = True,
                reverse_style = "until", reverse_data = 0):
        """ Here everthing the class needs is converted to a self varible of the class itself.

        Args:
            path (str): Path to the directory of the scans.
            tempPath (str): Path to the directory of the .metadata files.
            sampleName (str): Name of the sample.
            startScan (int): First scan that should be included.
            endScan (int): Last scan that should be included.
            timeOfset (int, optional): Offset for the plotted time. Defaults to 0.
            figratio1D (list, optional): Figure ratio for 1D plots. Defaults to [16,9].
            figratio2D (list, optional): Figure ratio for 2D plots. Defaults to [7, 4].
            reload (str, optional): Reload previously used data from .mat reload file. Defaults to None.
            average (int, optional): Averaging applied to the scans. Defaults to 30.
            without_dark (bool, optional): Excluding dark files again (if temperature and time is loaded from .metadata). Defaults to True.
            reverse_style (str, optional): PDF data can be reversed by "list" of every PDF index, 
                            "until" a certain index, "from a certain index", or "between" certain indices  . Defaults to "until".
            reverse_data (int, list, optional): list for style "list, float for "until" and "from", and list with size 2 for "between". Defaults to 0.

        """
        
        # Generate self varibles from input
        self.path = path
        self.sampleName = sampleName
        self.startScan = startScan
        self.endScan = endScan
        self.timeOfset= timeOfset
        self.without_dark = without_dark
        self.reload = reload
        self.average = average
        self.reverse_style = reverse_style 
        self.reverse_data = reverse_data
        
        self.max_all_int = 0
        self.min_all_int = 0

        # Normalize Ratios given
        ratio1D = figratio1D[0]/figratio1D[1] 
        figratio1D = [ratio1D*9,9]
        self.figratio1D = figratio1D
        ratio2D = figratio2D[0]/figratio2D[1] 
        figratio2D = [ratio2D*4,4]
        self.figratio2D = figratio2D

        # Check if earlier loaded data is given  
        if reload != None:
            
            # Here we use scipy for .mat data files
            loadedmat = sp_io.loadmat(reload)
            
            # .mat files are essentially like .csv or other tabulars, single arrays need [0], while matrixis do not
            self.G = loadedmat['G_of_r']
            self.r = loadedmat['r'][0]
            self.time_array = loadedmat['time'][0] + timeOfset
            self.temp_array = loadedmat['temp'][0]
            
            # We average dircetly from the whole set (reload data sets allways contain unaveraged data)
            self.time_averaged_array = loadedmat['time_averaged'][0][startScan:endScan] + timeOfset
            self.temp_averaged_array = loadedmat['temp_averaged'][0][startScan:endScan]
            
        else:
            # Import function handeles .gr 
            self.import_data()
            
            # Temperature and time are loaded from PDF generation reload files
            loadedmat = sp_io.loadmat(tempPath)
            self.time_array = loadedmat['time'][0] + timeOfset
            self.temp_array = loadedmat['temp'][0]
            self.time_averaged_array = loadedmat['time_averaged'][0][startScan:endScan] + timeOfset
            self.temp_averaged_array = loadedmat['temp_averaged'][0][startScan:endScan]
            
            # Averging is handelt driectly
            self.time_averaged_array = averageArray(self.time_array,self.average)
            self.temp_averaged_array = averageArray(self.temp_array,self.average)
            
            # Temperature and time is cropped to the same lenght as G(r)  
            self.temp_averaged_array = self.temp_averaged_array[:len(self.G)]
            self.time_averaged_array = self.time_averaged_array[:len(self.G)]

        # We select only the wanted G(r)
        self.G = self.G[startScan:endScan]
        
        # And if nessisary G(r) is flipped (G(r) = -G(r)). 
        # The scans to be flipped needs to be specified when declaring the class.
        self.reverse_PDF(self.reverse_style, self.reverse_data)
            


