"""

:platform: Unix, Windows 
:last changed: 2024-01-24

.. moduleauthor:: Tjark Leon Raphael Gröne <tgroene@physnet.uni-hamburg.de>


"""
import pathlib
rootdir = str(pathlib.Path(__file__).parent.resolve())

from time import time
from scipy import io as sp_io


from backbone.functions.plotting import Plotting_XRD
from backbone.functions.data_processing import Processing_Data
from backbone.functions.export import Export_PDF_Gen
from backbone.functions.import_data import Import_PDF_Gen
from backbone.functions.utils import averageArray

class PDF_generation_class(Plotting_XRD,Processing_Data,Export_PDF_Gen,Import_PDF_Gen):
    """ A class to load, process and plot XRD/TS data and transform it to PDFs.

    Args:
        Plotting_XRD (class): COntains plotting functions for XRD/TS data.
        Processing_Data (class): Contains functions to process XRD/TS data and fourier-transform it to PDF data. 
        Export_PDF_Gen (class): Contains functions to export data to .mat, .dat and .log files.
        Import_PDF_Gen (class): COntains functions to import .chi/.dat and .metadata files.
    """
    
    # The init function declairs all to be set varibles for the class.
    def __init__(self, path: str,tempPath: str,sampleName: str,startScan: int=0,endScan: int=1e10, timeOfset=0, ends_with = '.chi', without_dark=True, 
                figratio1D=[16,9], figratio2D=[7, 4], reload: str = None, number_of_steps: int=1, skiprows_of_files: int = 23, average: int=1, counter_PDF: dict=None,
                Composition: str="ZnS", Qmaxinst: float=30, Qmin: float=0.1, Qmax: float=30, Rpoly: float=0.9, Rstep: float=0.01, Rmin: float=0.5, Rmax: float=30):
        """ Class type object to automate the generation the PDF (.gr) from TS scans using PDFgetX3 within python.
            For in-situ data sets this class can combind TS scans and temperature readings as well as perform averaging binning, BG subtraticon, 
            exportation, filtering, interpolation, baseline fitting  and PDF generation.

        Args:
            path (str): Path to file
            tempPath (str): Path to the .metadata
            sampleName (str): Name of your sample (for Ledgend).
            startScan (int): First scan to include.
            endScan (int): Last scan to include.
            timeOfset (int, optional): Offset of the timeaxes for insitu measurements in s (-10 time axis starts with -10 s)  . Defaults to 0.
            ends_with (str, optional): file extension of the integrated data (.chi/.dat). Defaults to '.chi'.
            without_dark (bool, optional): Excluding dark detector images from  the run. Defaults to True.
            figratio1D (list, optional): Ratio of the plot window for 1D plots. Defaults to [16,9].
            figratio2D (list, optional): Ratio of the plot window for 2D plots. Defaults to [7, 4].
            reload (str, optional): Path to a saved MAT file, that was saved with the saveReload() function before. Defaults to None.
            number_of_steps (int, optional): load faster by introducing steps, that reduce data quality. Defaults to 1.
            skiprows_of_files (int, optional): header lines in integrated data. Defaults to 23.
            average (int, optional): Order of which scanns are binned and averaged over. Defaults to 1.
            Composition (str, optional): Moleculart formular of the material probed. Defaults to "ZnS".
            Qmaxinst (float, optional): Maximum Q that has physical meaning (excluding mask, etc.). Defaults to 30.
            Qmin (float, optional): Minuimum Q to include in the PDF generation. Defaults to 0.1.
            Qmax (float, optional): Maximum Q to include in the PDF generation. Defaults to 30.
            Rpoly (float, optional): Polynomal degree of data correction applied (best case: rpoly < Qmin). Defaults to 0.9.
            Rstep (float, optional): Step with of the PDF (Smaller steps result in longer calculation times). Defaults to 0.01.
            Rmin (float, optional): Minimum r the PDF is calculated from. Defaults to 0.5.
            Rmax (float, optional): Maximum r the PDF is calculated to. Defaults to 30.
            
        References:
            PDFgetX3:
                Code available from https://www.diffpy.org/products/pdfgetx.html
                Related Paper:  P. Juhás and T. Davis, C. L. Farrow, S. J. L. Billinge,https://journals.iucr.org/j/issues/2013/02/00/nb5046/index.html
                                J. Appl. Crystallogr. 46, 560-566 (2013)
                     
        """
        self.t0 = time() #get starting time 

        # Conversion to self varibles
        self.path = path
        self.sampleName = sampleName
        self.startScan = int(startScan)
        self.endScan = int(endScan)
        self.timeOfset= timeOfset
        self.without_dark = without_dark
        self.reload = reload
        self.number_of_steps = number_of_steps
        self.ends_with = ends_with
        self.skiprows_of_files = skiprows_of_files
        self.average = average
        self.glas_Gr = False
        self.Composition = Composition
        self.Qmaxinst = Qmaxinst
        self.Qmin = Qmin
        self.Qmax = Qmax
        self.Rpoly = Rpoly
        self.Rstep = Rstep 
        self.Rmin = Rmin
        self.Rmax = Rmax
        self.counter_PDF = counter_PDF
        
        self.max_all_int = 0
        self.min_all_int = 0

        # Normalize Ratios given
        ratio1D = figratio1D[0]/figratio1D[1] 
        figratio1D = [ratio1D*9,9]
        self.figratio1D = figratio1D
        ratio2D = figratio2D[0]/figratio2D[1] 
        figratio2D = [ratio2D*4,4]
        self.figratio2D = figratio2D

        # Save the initial params to .log file.
        self.logger = {"Initial parameters":{   "Path to .chi/.dat": path,
                                                "Path to .metadata": tempPath,
                                                "File Extension of integrated data": ends_with,
                                                "Header lines skipped": skiprows_of_files,
                                                "Data was loop over in number of steps": number_of_steps,
                                                "Sample name": sampleName,
                                                "Start scan": startScan,
                                                "End scan": endScan,
                                                "time Ofset": timeOfset,
                                                "Data was reloaded": self.reload,},}

        
        # Certain varibles can be stored in a .mat file to reload data much faster next time and make reprocessing timesmuch faster.
        if reload != None:
            
            # Here we use scipy for .mat data files
            loadedmat = sp_io.loadmat(reload)
            
            # .mat files are essentially like .csv or other tabulars, single arrays need [0], while matrixis do not
            self.I = loadedmat['I_matrix']
            self.q = loadedmat['q_axis'][0]
            self.time_array = loadedmat['time'][0] + timeOfset
            self.temp_array = loadedmat['temp'][0]
            
            # Checking for already declaired averaged data 
            self.time_averaged_array = loadedmat['time_averaged'][0][self.startScan:self.endScan] + timeOfset
            self.temp_averaged_array = loadedmat['temp_averaged'][0][self.startScan:self.endScan]
            
            
            self.number_of_files = len(self.I)
            
            
            # We average dircetly from the whole set (reload data sets allways contain unaveraged data)
            self.time_averaged_array = averageArray(self.time_array,self.average)
            self.temp_averaged_array = averageArray(self.temp_array,self.average)
            self.useAverage()
        
        else:
            # If this is the first time the data is loaded, the program imports data and metadata and averages.
            self.import_data()
            self.getTempTime(tempPath,self.counter_PDF)
            self.useAverage()

        # Giving loading time feedback in the consol
        print("\n Time for processing")
        print("------------------------------------------------------------------")
        print("Time: " + str((time() - self.t0)/60) + "\n")
        
        # Save processing time in the .log file.
        self.logger['Time it took to load'] = {'Time in min': str((time() - self.t0)/60)}





