import numpy as np    
import pathlib
import os
import subprocess
from scipy.signal import savgol_filter
from scipy import interpolate
from backbone.functions.utils import *
from backbone.functions.export import outputDatFile
from pybaselines import Baseline
from backbone.functions.whittaker_smooth import whittaker_smooth
import pathlib
import random

rootdir = str(pathlib.Path(__file__).parent.resolve())

class Processing_Data:
    """ Class to filter, average, rebin, BG-subtract and generate PDFs from XRD/TS data.

    Raises:
        Exception: _description_
    """
    #----------------------------------------------------------------------------------------#
    #                                 Get Start and End of ramp                              #
    #----------------------------------------------------------------------------------------#
    def getTempPoints(self,rampTemp,holdTemp):
        """ 
        
        Arguments
        ----------
        rampTemp:   temperature the ramp is defined to start 
        holdTemp:   temperature the plato is defined to start
        """
        
        # Logging the event
        self.logger['Scan idizes of the ramp satrt and end collected'] = {  'Temperature the ramp is set to start': rampTemp,
                                                                            'Temperatrue the plato is set to start': holdTemp}
        
        # Set initial status varibles
        ramp_started = False
        plato_started = False
        plato_ended = False

        # Normalize the file index for start and end scan given 
        if self.startScan > 0:
            if self.endScan < self.number_of_files*self.number_of_steps:
                file_index = np.linspace(self.startScan,self.endScan,self.number_of_files)
            else:
                file_index = np.linspace(self.startScan,self.number_of_files*self.number_of_steps,self.number_of_files)
        else: 
            if self.endScan < self.number_of_files*self.number_of_steps:
                file_index = np.linspace(1,self.endScan,self.number_of_files)
            else:
                file_index = np.linspace(1,self.number_of_files*self.number_of_steps,self.number_of_files)

        # Give feedback to the user about the status of the integration. 
        # --> If more .metadata files are available the .chi/.dat.
        if len(file_index) < len(self.temp_array):   
            print('\n\n ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
            print('          Live integration lags behind by ' + str(len(self.temp_array) - len(file_index)) + ' scans')
            print(' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n')

        # Here we loop over the temperature array
        for i,j in enumerate(self.temp_array):
            
            # And check if this is still below the temperature threshold of the ramp.
            if j > rampTemp and ramp_started==False:
                
                # If the threshold is reached, this experimantal temperature is logged and
                # the status varible is updated
                print('Ramp start: ', file_index[i],'Temp: ', j)
                self.rampstart_index = int(file_index[i])
                ramp_started = True 
                self.logger['Ramp Start'] = {   'Index': file_index[i],
                                                'Temperature': j}
                
            # Here we check if the ramp has finished and the final temperature is reached
            if j > holdTemp and plato_started==False:
                
                # If the threshold is reached, this experimantal temperature is logged and
                # the status varible is updated
                print('Plato start: ',file_index[i],'Temp: ', j)
                self.platostart_index = int(file_index[i])
                plato_started = True 
                self.logger['Plato Start'] = {  'Index': file_index[i],
                                                'Temperature': j}
            
            # Finally we check when the cooldown is reached
            if j < holdTemp-5 and plato_ended==False and plato_started==True:
                
                # If the threshold is reached, this experimantal temperature is logged and
                # the status varible is updated
                print('Plato endet: ', file_index[i],'Temp: ', j)
                plato_ended = True
                self.platoend_index = int(file_index[i]) 
                self.logger['Plato End'] = {   'Index': file_index[i],
                                                'Temperature': j}
        
        # If a threshold is never reached an error is risen.
        if 'platostart_index' not in dir(self):
            raise Exception("Temperature always below the set tempertaure defined") 
        
        # If no cooldown is measured we stop
        if 'platoend_index' not in dir(self):
            self.platoend_index = len(self.temp_array)
        return self.logger, self.rampstart_index, self.platoend_index, self.platoend_index

    #----------------------------------------------------------------------------------------#
    #                              Shorten in Q and Normalize Data                           #
    #----------------------------------------------------------------------------------------#
    def norm(self, normalize: str=None):
        """ Function to norm XRD/TS data.
        
        Arguments
        ----------
        normalize:  normalisation mode None, max or minmax
        Qmaxinst:   cut q to better fit the background subtration late
        """
        
        # Logging the event
        self.logger['Data was normalized and or cutted in q'] = {   'Normalization mode': normalize}
        
        # Set temporary varibles
        q = None
        new_I = []
        
        # Loop over each Scan
        for i,j in enumerate(self.I):
            
            # Norm just to the maximum of each scan setting it to 1.
            if normalize == 'max':
                I_norm = I_norm/max(I_norm)
                
            # Norm the maximum to 1 and the minimun to 0. 
            if normalize == 'minmax':       
                I_norm = (I_norm - min(I_norm)) / (max(I_norm)-min(I_norm))

            # create new matrix
            new_I.append(I_norm)
        
        # update self varible
        self.I = new_I
        return self.logger, self.I, self.q



    #----------------------------------------------------------------------------------------#
    #                                          Average                                       #
    #----------------------------------------------------------------------------------------#
    def useAverage(self):
        """ Averages the I(q) data over a set of scans
        
        Arguments
        ----------
        average:    number of files to average over
        """
        
        #Storing Event in logger
        self.logger['Data was averaged'] = {   'Averaged over': self.average}
        
        #Create copie of I(q), thus it can be accesed later on
        self.I_original = self.I
        
        # Average errors if exsiting
        # self.dI_original = self.dI_matrix

        
        # Update I(q) (and dI(q) to the averaged data
        self.I = averageArray(self.I,self.average,2)
        # self.dI = averageArrayError(self.dI,self.average,2)
                
        
        return self.I, self.I_original  #, self.dI ,self.dI_original 

    
    #----------------------------------------------------------------------------------------#
    #                               Get parts of ramps and platos                            #
    #----------------------------------------------------------------------------------------#
    def getRampPlatoLists(self,heat_ranges):
        """ Check which parts of the data belong to a plato (befor starting the temperature ramp and when holding the final temperature)
            and which parts belong to a temperature ramp.
        
        Arguments
        ----------
        heat ranges:    heat ranges dictionary including the start and end of the temp ramp
        """
        
        # Logging the event
        self.logger['I(q) for Ramp and Plato where collected in lists'] = {   'Heatrages input': heat_ranges,}
        
        # Create the empty lists
        plato_file_list = []
        self.plato_list = [[],[]]
        ramp_file_list = []
        self.ramp_list = [[],[]]

        # Step one is the plato at room temperature befor ramping
        step = 1
        
        # Give user feedback about the determined temperatures 
        print(heat_ranges)
        
        # Loop over every I(q) from each scan
        for i, j in enumerate(self.I):
            
            # If step is odd it´s a plato
            if step % 2 != 0:
                plato_file_list.append(j)
            
            # If step is even it´s ramp
            if step % 2 == 0:
                ramp_file_list.append(j)

            # Here we check when the ramp starts and account for averaging
            if  i == round_to_multiple(heat_ranges["ramp1_start"],self.average) /self.average:
                
                # We put the array of all I(q) from the first plato (at room temperature) as a first element of a list of all platos
                self.plato_list[0].append(plato_file_list)
                plato_file_list = []
                
                # We move on to the first ramp 
                step = 2

            # Here we check when the first set temperature starts and account for averaging
            if i == round_to_multiple(heat_ranges["ramp1_end"],self.average) /self.average:
                
                # We put the array of all I(q) from the first ramp as a first element of a list of all ramps
                self.ramp_list[0].append(ramp_file_list)
                ramp_file_list = []
                step = 3

            # Here we check when second ramp starts (if set) and account for averaging
            if i == round_to_multiple(heat_ranges["ramp2_start"],self.average) /self.average:
                
                # We put the array of all I(q) from the second plato (at the first set temperature) as a first element of a list of all platos
                self.plato_list[0].append(plato_file_list)
                plato_file_list = []
                
                # It´s rare but we also can move on to a second ramp 
                step = 4

            # Here we check when the second set temperature (if set) starts and account for averaging
            if i == round_to_multiple(heat_ranges["ramp2_end"],self.average) /self.average:
                
                # We put the array of all I(q) from the second ramp as a first element of a list of all ramps
                self.ramp_list[0].append(ramp_file_list)
                ramp_file_list = []
                
                # It´s rare but we also can move on to a second hold tmperature
                step = 5

        # It´s rare but we also can put the array of all I(q) from the third plato (at the second set temperature) as a first element of a list of all platos
        self.plato_list[0].append(plato_file_list)
        plato_file_list = []
        
        return self.logger, self.plato_list, self.ramp_list


    #----------------------------------------------------------------------------------------#
    #                                  Background Subtraction                                #
    #----------------------------------------------------------------------------------------#
    def bgSubtraction(self,plato_bg_list, ramp_bg_list, heat_ranges, bg_sampleName, outputPath, mode: str='last', 
                      Qmaxinst: float=30, FactorMax: float=0.99,single_BG_file: str=None, empty_glas_int: float=0,
                      empty_glas: str=None, Qmin: float=None, Qmax: float=None, use_PDF_maximization: bool=False):
        """Automated BG subtraction 

        Args:
            plato_bg_list (list): list of array from the plato related background scans.
            ramp_bg_list (list): list of array from the ramp related background scans.
            heat_ranges (dict): heat ranges dictionary including the start and end of the temp ramp.
            bg_sampleName (str): Name of the BG sample for .dat output.
            outputPath (str): Path to where .dat generated after BG subtration are saved. 
            mode (str, optional): Choose "last" to repeat the last background scan or "average" to use an everage of the plato. Defaults to 'last'.
            FactorMax (float, optional): Maximum factor allowed for the background intesity subtracted from the data. Defaults to 0.99.
            single_BG_file (str, optional): Path to a single background if only one background file should be subtracted from all data. Defaults to None.
            empty_glas_int (float, optional): Intensity with which a constant contribution of glas is removed from each scan. Defaults to 0.
            empty_glas (str, optional): Path to the empty glas file. Defaults to None.
            Qmin (float, optional): Minimum Q that is included in the BG subtrastion (this can purposly differ from Qmin for PDF generation). Defaults to None.
            Qmax (float, optional): Maximum Q that is included in the BG subtrastion (this can purposly differ from Qmin for PDF generation). Defaults to None.
            use_PDF_maximization (bool, optional): Make a rought PDF around the expected Peak and maximize the PDF iterativ. Defaults to False. 

        Returns:
            self: updated self varibles
        """
        
        self.flipps = 0
        self.alltime_low = 1e10
        self.alltime_low_area = 1e10
        self.alltime_low_area2 = 1e10
        
        # Logging the event
        self.logger['Background was subtracted'] = {'heat ranges used': heat_ranges,
                                                    'Name of the background': bg_sampleName,
                                                    'Mode used for shorter background array': mode,
                                                    'Maximum factor allowed': FactorMax,
                                                    'Path to single background file (if used)': single_BG_file,
                                                    'Q cutted at Qmaxinst': Qmaxinst,}
        
        
        # Cropping the data in q before subtraction (mask affected data), so it does not form artefacts in the data
        clostest_Qmaxinst = find_nearest(self.q,Qmaxinst)
        clostest_Qmaxinst = int(np.where(self.q==clostest_Qmaxinst)[0])
        self.clostest_Qmaxinst = clostest_Qmaxinst
        self.q = self.q[:clostest_Qmaxinst]
        shorten_matrix_elements(self.I,[0,clostest_Qmaxinst])
        
        print('\n\n Background factors and ')
        
        # FInding the corresponding Qmax and Qmin in the q array, or which number is closest to it and getting the index
        if Qmax != None:
            Qmax = find_nearest(self.q,Qmax)
            Qmax = int(np.where(self.q==Qmax)[0][0])
        else:
            Qmax =-1
        if Qmin != None:
            Qmin = find_nearest(self.q,Qmin)
            Qmin = int(np.where(self.q==Qmin)[0][0])
        else:
            Qmin = 0
            
        # Loading empty glas contrebution if given
        if empty_glas != None: 
            self.q_empty_glas, self.I_empty_glas = np.loadtxt(empty_glas,skiprows=1).T
        else:
            self.q_empty_glas = 0
            self.I_empty_glas = 0
        
        # Setting new lists for factors and platos and ramps
        self.plato_bg_list = plato_bg_list
        self.ramp_bg_list = ramp_bg_list
        self.factor_list = []
        self.raw_factor_list = []
        self.subtracted_I = []
        
        # If the BG is temperature or time dependent one has two options if the data set is shorter then the real experiment 
        if single_BG_file == None:
            
            # The first option is to use the last scan over and over
            if mode == 'last':
                
                # Then we append the last scan as many times as scans are missing to mich the real experiment
                for i,j in enumerate(self.plato_list[0]):
                    if len(j) > len(self.plato_bg_list[0][i]):
                        for k in range(int(len(j)-len(self.plato_bg_list[0][i]))):
                            self.plato_bg_list[0][i].append(self.plato_bg_list[0][i][-1])
                            
            # Or one can average over the final temperature plato
            # However the scale is often changing and I recoment using the first option
            if mode == 'average':
                average = 0
                average_list = []
                
                # First we load all data files and add them up
                for i,j in enumerate(self.plato_bg_list[0]): 
                    for a,b in enumerate(j): 
                        average = average + getIfile(b)
                        
                    # Then we devide by the number of all scans forming the overall average
                    average = average/len(j)
                    
                    # Then we append the averaged scan as many times as plato scans of the final temperature exsist
                    average_list.append(average)
                    average = 0

        # We give the user feedback about the leght of the platos and ramps
        # print('Lenght of 1st Plato BG:  ' + str(len(self.plato_bg_list[0][0])),'    Lenght of 2nd Plato BG:  ' + str(len(self.plato_bg_list[0][1])), '    Lenght of 1st Ramping BG:  ' + str(len(self.ramp_bg_list[0][0])))
        # print('Lenght of 1st Plato Measurement:  ' + str(len(self.plato_list[0][0])),'    Lenght of 2nd Plato Measurement:  ' + str(len(self.plato_list[0][1])), '    Lenght of 1st Ramping Measurement:  ' + str(len(self.ramp_list[0][0])))
        
        # # 
        # self.bg_matrix = self.plato_bg_list[0][0] + self.ramp_bg_list[0][0] + self.plato_bg_list[0][1]
        self.bg_matrix = self.plato_bg_list[0][0] + self.ramp_bg_list[0][0] + self.plato_bg_list[0][1]
        
        
        # 
        if self.glas_Gr == True:
            self.bg_matrix_Gr = self.plato_bg_list[1][0] + self.ramp_bg_list[1][0] + self.plato_bg_list[1][1]
        
        # Cropping the BG data as well in q before subtraction (mask affected data), so it does not form artefacts in the data
        shorten_matrix_elements(self.plato_bg_list[0][0],[0,clostest_Qmaxinst])
        shorten_matrix_elements(self.plato_bg_list[0][1],[0,clostest_Qmaxinst])
        shorten_matrix_elements(self.ramp_bg_list[0][0],[0,clostest_Qmaxinst])
        shorten_matrix_elements(self.bg_matrix,[0,clostest_Qmaxinst])
        
        # If one single BG is sufficient one can use this as well (a long exposed or averaged file works nicely)
        if single_BG_file != None:
            
            # We get the single background file
            self.I_bg = getIfile(single_BG_file)
            
            # Then we loop over all I(q)
            for i,j in enumerate(self.I):  
                
                # We can determine the BG scaling factor eighter by the intensity of the glas peak in the PDF or
                # by letting the minimum of the scan to be at the lowest point to be exactly 0 (within the given Q range)
                if self.glas_Gr == True:
                    self.factor = max(self.Gr[i]) / max(self.Gr_bg)
                    print(self.factor)
                    print('max Gr:  ' + str(max(self.Gr[i])))
                    print('max Gr bg:  ' + str(max(self.Gr_bg)))                    
                
                # If the factor is not determined from the glas peak intensity (for some organic it is highly overshadowed)
                else:
                    self.I_bg = self.I_bg[0:clostest_Qmaxinst]
                    I_bg_frac = j[Qmin:Qmax] / self.I_bg[Qmin:Qmax] 
                    self.factor = min(I_bg_frac)
                
                # Check if the calculated factor is over the allowed limit, otherwise it is reseted
                if FactorMax != None:
                    if self.factor > FactorMax:
                        self.factor = FactorMax
                
                # Give user feedback about the current factor        
                print(self.factor)

                # Aplly the current factor to the data
                self.I[i] = j - self.factor*self.I_bg
                
                # ????
                if min(self.I[i]) < 0:
                    self.I[i] = j - self.factor*self.I_bg - min(self.I[i])
                    
                    
                if i ==0:
                    outputDatFile(rootdir + '/../../processed/' +self.sampleName +'/Dat_files/bg_type1_uncorrected_monte_Dat/',self.q,[self.I_bg],(i+1)*self.average,'background_'+self.sampleName, clear_folder=True)
                else:
                    outputDatFile(rootdir + '/../../processed/' +self.sampleName +'/Dat_files/bg_type1_uncorrected_monte_Dat/',self.q,[self.I_bg],(i+1)*self.average,'background_'+self.sampleName)
                
                # After creating the .dat of the BG the PDFs can be calculated in monte carlo style 
                # for maximization of the G(r)
                if use_PDF_maximization == True:
                    self.factor, maxvalue = self.monte_carlo_PDF_maximization(rootdir + '/../../processed/' + self.sampleName + '/Dat_files/meas_type1_uncorrected_Dat/' + self.sampleName + "_" + str((i+1)*self.average).zfill(5)+".txt", 
                                                    rootdir + '/../../processed/' + self.sampleName + '/Dat_files/bg_type1_uncorrected_monte_Dat/' + 'background_' + self.sampleName + "_" + str((i+1)*self.average).zfill(5)+".txt",outputPath, scan_number=i)
                                                    #r"C:\Users\admin\Nextcloud\Python_Code\ProjectPDF\PDF_generation\From_server\PDF_generation\processed\cell2_fe3s4_in_bnoh_160c_60min\empty_glasinlet_0p3m-00001.dat",outputPath)
                    print("Flipps done:" + str(self.flipps)) 
        
        else:
            
            #
            I = 0
            
            # Counter for the files in a ramp or plato processed already
            count = 0
            # Offset between the number of room temperature files between BG and experimantal run
            self.bg_offset = 0
            
            # If the BG scans have less data points then the experimantal run we start later in the experimantal run,
            # by setting the count to when the offset between the two data sets
            # --> Thereby the earliest data is ignored as we don´t have enought BG files 
            #     (it would be better to treat it the same way as the end of run data, but mostly it should´t differ so I did´t include it so fare)
            if len(self.plato_bg_list[0][0]) > (round_to_multiple(heat_ranges["ramp1_start"],self.average)/self.average):
                self.bg_offset = int(len(self.plato_bg_list[0][0]) - (round_to_multiple(heat_ranges["ramp1_start"],self.average)/self.average) - 1)
                print('BG is offset by:  ' + str(self.bg_offset) + '  number of averaged scans')
                count = self.bg_offset
            
            # Steps are used in the same manner as in getRampPlatoLists (step 1 is the room temperature plato)
            step = 1
            
            # Now we can loop over every I(q) from each scan in the experimantel run
            for i,j in enumerate(self.I):  
                self.current_I = j
                # Here we check when the ramp starts and account for averaging
                if i > round_to_multiple(heat_ranges["ramp1_start"],self.average)/self.average and step < 2:
                    step = 2
                    # We reset the counter after each step
                    count = 0

                # Here we check when the first set temperature starts and account for averaging
                if i-1 == round_to_multiple(heat_ranges["ramp1_end"],self.average)/self.average:
                    step = 3
                    # We reset the counter after each step
                    count = 0

                # Here we check when second ramp starts (if set) and account for averaging
                if i-1 == round_to_multiple(heat_ranges["ramp2_start"],self.average)/self.average:
                    step = 4
                    # We reset the counter after each step
                    count = 0

                # Here we check when the second set temperature (if set) starts and account for averaging
                if i-1 == round_to_multiple(heat_ranges["ramp2_end"],self.average)/self.average:
                    step = 5
                    # We reset the counter after each step
                    count = 0

                # The ramps should have the same nbumber of files in  BG and experimantel run, so they are not so complicated
                if step == 2:
                    
                    # If better implemented this doesn´t need to be a try function :(
                    try:
                        
                        # Reshape the order of the BG to match the experimantel temperature curve
                        self.I_bg =  self.ramp_bg_list[0][0][count]
                        
                        if self.glas_Gr == True:
                            self.Gr_bg = self.ramp_bg_list[1][0][count]
                        
                        # Give the user feedback what scan is in process and where in the run it is
                        print('Scan:  ' + str(count),'    in Ramping with lenght: ' + str(len(self.ramp_bg_list[0][0])-1), '   step in procedure:  ' + str(step))
                    
                    except:
                        # This should be taken over by the automation 
                        # (if averaging and temperature differneces between BG and experimantel run make up for a unfortuned 1 scan differece in the end)
                        print('optimize step 2')

                if step == 4:
                    
                    # If better implemented this doesn´t need to be a try function :(
                    try:
                        
                        # Reshape the order of the BG to match the experimantel temperature curve
                        self.I_bg = self.ramp_bg_list[0][1][count]
                        if self.glas_Gr == True:
                            self.Gr_bg = self.ramp_bg_list[1][1][count]
                         
                        # Give the user feedback what scan is in process and where in the run it is
                        print('Scan:  ' + str(count),'    in Ramping with lenght: ' + str(len(self.ramp_bg_list[0][1])-1), '   step in procedure:  ' + str(step))
                    except:
                        # This should be taken over by the automation 
                        # (if averaging and temperature differneces between BG and experimantel run make up for a unfortuned 1 scan differece in the end)
                        print('optimize step 4')

                # The platos will most of the time have diffent numbers of files in BG and experimantel run, so we use the extended versions of the BG run we created in the beginning
                # with last copies the last scan until it matches the experimantal run leght or average which uses an average of the BG plato and extends it
                if mode == 'last':
                    
                    # If better implemented this doesn´t need to be a try function :(
                    if step == 1:
                        
                        # Reshape the order of the BG to match the experimantel temperature curve
                        self.I_bg =  self.plato_bg_list[0][0][count]
                        if self.glas_Gr == True:
                            self.Gr_bg =  self.plato_bg_list[1][0][count]
                        
                        # Give the user feedback what scan is in process and where in the run it is
                        print('Scan:  ' + str(count),'    in Plato with lenght: ' + str(len(self.plato_bg_list[0][0])-1), '   step in procedure:  ' + str(step))

                    # If better implemented this doesn´t need to be a try function :(
                    if step == 3:
                        try:
                            
                            # Reshape the order of the BG to match the experimantel temperature curve
                            self.I_bg = self.plato_bg_list[0][1][count]
                            if self.glas_Gr == True:
                                self.Gr_bg = self.plato_bg_list[1][1][count]
                            
                             # Give the user feedback what scan is in process and where in the run it is
                            print('Scan:  ' + str(count),'    in Plato with lenght: ' + str(len(self.plato_bg_list[0][1])-1), '   step in procedure:  ' + str(step))
                        except:
                            # This should be taken over by the automation 
                            # (if averaging and temperature differneces between BG and experimantel run make up for a unfortuned 1 scan differece in the end)
                            print('optimize step 3')

                    # If better implemented this doesn´t need to be a try function :(
                    if step == 5:
                        try:
                            
                            # Reshape the order of the BG to match the experimantel temperature curve
                            self.I_bg = self.plato_bg_list[0][2][count]
                            if self.glas_Gr == True:
                                self.Gr_bg = self.plato_bg_list[1][2][count]
                            
                            # Give the user feedback what scan is in process and where in the run it is
                            print('Scan:  ' + str(count),'    in Plato with lenght: ' + str(len(self.plato_bg_list[0][2])-1), '   step in procedure:  ' + str(step))
                        except:
                            # This should be taken over by the automation 
                            # (if averaging and temperature differneces between BG and experimantel run make up for a unfortuned 1 scan differece in the end)
                            print('optimize step 5')

                # This is for the average of the BG plato that is extended, here the BG platos are simply the average of the current plato
                if mode == 'average':
                    if step == 1:
                        self.I_bg = average_list[0]
                        
                    if step == 3:
                        self.I_bg = average_list[1]
                        
                    if step == 5:
                        self.I_bg = average_list[2]
                
                # Here we subtract the empty glas file with a constant scale before determaning the factor for the subtraction
                if empty_glas_int != 0:  
                    # Here we also shorten the leght to be matching to Qmaxinst
                    self.I_empty_glas = self.I_empty_glas[:len(j)]  
                    j = j - self.I_empty_glas*max(j)*empty_glas_int
            
                # We can determine the BG scaling factor eighter by the intensity of the glas peak in the PDF or
                # by letting the minimum of the scan to be at the lowest point to be exactly 0 (within the given Q range)
                if self.glas_Gr == True:
                    self.factor = max(self.Gr[i]) / max(self.Gr_bg)
                    print(self.factor)
                    print('max Gr:  ' + str(max(self.Gr[i])))
                    print('max Gr bg:  ' + str(max(self.Gr_bg)))                    
                
                # If the factor is not determined from the glas peak intensity (for some organic it is highly overshadowed)
                else:
                    I_bg_frac = j[Qmin:Qmax] / self.I_bg[Qmin:Qmax] 
                    self.factor = min(I_bg_frac)
                
                # Check if the calculated factor is over the allowed limit, otherwise it is reseted
                if FactorMax != None:
                    if self.factor > FactorMax:
                        self.factor = FactorMax
                
                # Here we apply the BG factor and subtract the BG from the experimantal I(q)
                subtracted_I = j - self.factor*self.I_bg
            
                
                # For the creation of the PDFs, which are needed for the monte carlo BG subtraction,
                # we need the BG in the correct order (ramp-BG with the same time as ramp-experimental)
                # Therefore we now create the .dat after sorting it
                
                # If we reprocess with different averages we need to clear the folder, therefore we do this for the first scan.
                if i ==0:
                    outputDatFile(rootdir + '/../../processed/' +self.sampleName +'/Dat_files/bg_type1_uncorrected_monte_Dat/',self.q,[self.I_bg],(i+1)*self.average,'background_'+self.sampleName, clear_folder=True)
                else:
                    outputDatFile(rootdir + '/../../processed/' +self.sampleName +'/Dat_files/bg_type1_uncorrected_monte_Dat/',self.q,[self.I_bg],(i+1)*self.average,'background_'+self.sampleName)
                
                # After creating the .dat of the BG the PDFs can be calculated in monte carlo style 
                # for maximization of the G(r)
                if use_PDF_maximization == True:
                    self.factor, maxvalue = self.monte_carlo_PDF_maximization(rootdir + '/../../processed/' + self.sampleName + '/Dat_files/meas_type1_uncorrected_Dat/' + self.sampleName + "_" + str((i+1)*self.average).zfill(5)+".txt", 
                                                    rootdir + '/../../processed/' + self.sampleName + '/Dat_files/bg_type1_uncorrected_monte_Dat/' + 'background_' + self.sampleName + "_" + str((i+1)*self.average).zfill(5)+".txt",outputPath, scan_number=i)
                                                    #r"C:\Users\admin\Nextcloud\Python_Code\ProjectPDF\PDF_generation\From_server\PDF_generation\processed\cell2_fe3s4_in_bnoh_160c_60min\empty_glasinlet_0p3m-00001.dat",outputPath)
                    print("Flipps done:" + str(self.flipps)) 
                    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
                    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#  
                    # Here is my experimantal corner triing out FT low pass and notch filters, making PDFs from the empty cell, etc...
                    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
                    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
                    
                     
                    # self.factor, self.glas_factor, maxvalue = self.monte_carlo_PDF_maximization_with_glas_removal(rootdir + '/../../processed/' + self.sampleName + '/Dat_files/meas_type1_uncorrected_Dat/' + self.sampleName + "_" + str((i+1)*self.average).zfill(5)+".txt", 
                    #                                         rootdir + '/../../processed/' + self.sampleName + '/Dat_files/bg_type1_uncorrected_monte_Dat/' + 'background_' + self.sampleName + "_" + str((i+1)*self.average).zfill(5)+".txt",empty_glas,outputPath)
                    #                                         #r"C:\Users\admin\Nextcloud\Python_Code\ProjectPDF\PDF_generation\From_server\PDF_generation\processed\zns_2p_10_2\Dat_files\empty_ac_cell_inlet_180_10_Av120.txt", empty_glas,outputPath)
                    
                    # print(self.glas_factor)
                
                    # import matplotlib.pyplot as plt
                    # if i > 30:
                    #     # plt.plot(self.q,j)
                    #     # plt.plot(self.q,self.factor*self.I_bg)
                        
                    #     from scipy import fftpack
                    # I_minus_bg = j-self.factor*self.I_bg
                    # I_fft = fftpack.rfft(I_minus_bg)
                    # sample_freq = fftpack.fftfreq(I_minus_bg.size, d=self.q[1]-self.q[0])
                    # high_freq_fft = I_fft.copy()
                    # high_freq_fft[sample_freq > 7.5 ] = 0
                    # # high_freq_fft[sample_freq < 1.5 ] = 0
                    # filtered_I = fftpack.irfft(high_freq_fft)
                        
                    #     plt.plot(self.q,j-self.factor*self.I_bg)
                    #     plt.plot(self.q,filtered_I)
                    #     # plt.plot(self.q,j-self.factor*self.I_bg-self.glas_factor*self.I_empty_glas[:len(j)])
                    #     plt.xlim(left=0.5)
                    #     plt.ylim(bottom=0)
                    #     plt.show()


                else:
                    # The user gets the feedback about the calculated BG scale
                    print('BG_scale:  ' + str(self.factor))

                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
                #----- This part is about trying better BG corrections with baseline correction: -----#
                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
                # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#
                
                # +++++ Very Experimental +++++ #
                # turns = 0
                # while min(subtracted_I[Qmin:Qmax]) < 0 and turns < 1000:
                #     min_index = np.where(min(subtracted_I[Qmin:Qmax])==subtracted_I[Qmin:Qmax])[0][0]
                #     self.factor =  j[int(Qmin+min_index)] / self.I_bg[int(Qmin+min_index)]
                #     #subtracted_I[Qmin:Qmax] = subtracted_I[Qmin:Qmax] - min(subtracted_I[Qmin:Qmax])
                #     subtracted_I = j - self.factor*self.I_bg 
                #     turns = turns + 1
                
                # self.raw_factor_list.append(self.factor)
                
                # sub_baseline = 0
                # subtracted_I = subtracted_I - sub_baseline
                
                # turns = 0
                # while min(subtracted_I[Qmin:Qmax]) < 0 and turns < 1000:
                #     min_index = np.where(min(subtracted_I[Qmin:Qmax])==subtracted_I[Qmin:Qmax])[0][0]
                #     self.factor =  j[int(Qmin+min_index)] / self.I_bg[int(Qmin+min_index)]
                #     #subtracted_I[Qmin:Qmax] = subtracted_I[Qmin:Qmax] - min(subtracted_I[Qmin:Qmax])
                #     subtracted_I = j - self.factor*self.I_bg - sub_baseline
                #     turns = turns + 1

                # j = j -sub_baseline 
                # subtracted_I = np.clip(subtracted_I,0.001,None)
        
                # min_list = [l for l,v in enumerate(subtracted_I) if v == 0.001]
                # for l in min_list:
                #     j[l] = self.I_bg[l]*self.factor
                # j[:Qmin] = self.I_bg[:Qmin]*self.factor
                
                # if i ==0:
                #     outputDatFile(rootdir + '/../../processed/' +self.sampleName +'/Dat_files/meas_type3_corrected_after_Dat/',self.q,[j],i+1*self.average,self.sampleName, clear_folder=True)
                #     outputDatFile(rootdir + '/../../processed/' +self.sampleName +'/Dat_files/bg_type3_corrected_after_Dat/',self.q,[self.I_bg],i+1*self.average,'background_'+self.sampleName, clear_folder=True)
                # else:
                #     outputDatFile(rootdir + '/../../processed/' +self.sampleName +'/Dat_files/meas_type3_corrected_after_Dat/',self.q,[j],i+1*self.average,self.sampleName)
                #     outputDatFile(rootdir + '/../../processed/' +self.sampleName +'/Dat_files/bg_type3_corrected_after_Dat/',self.q,[self.I_bg],i+1*self.average,'background_'+self.sampleName)
                # # self.outputMatFile(rootdir + '/processed/' + self.sampleName + '/Mat_files/meas_type3_corrected_after_Mat/')
                
                
                # We store the BG subtracted I(q) and the BG scaling factor in self varibles for later access.
                self.subtracted_I.append(subtracted_I)
                self.factor_list.append(self.factor)
                
                # AT last the counter for each scans in each step is increased by 1.
                count = count + 1

        return self.logger,self.current_I, self.factor, self.raw_factor_list, self.factor_list, self.I_bg, self.bg_matrix, self.subtracted_I, self.plato_bg_list[0], self.ramp_bg_list[0], self.q_empty_glas, self.I_empty_glas, self.bg_offset
    
    
    #----------------------------------------------------------------------------------------#
    #                                   Correct for Baseline                                 #
    #----------------------------------------------------------------------------------------#
    def baselineCorrection(self,max_half_window):
        """ Fits a SNIP baseline to the data and subtracts it.
        
        Arguments
        ----------
        max_half_window:    Window size of the applied SNIP baseline correction
        """
        
        # Logging the event
        self.logger['Baseline was corrected for'] = {'Max of the half window': max_half_window,}
        
        # Set X-Axis for the baseline
        baseline_fitter = Baseline(x_data=self.q)
        
        # Set empty list for processed data
        self.baseline = []
        
        # Loop over every I(q) from each scan
        for i, j in enumerate(self.I):
            
            # Calculate the baseline and subtract it from data
            self.I[i] = self.I[i]- baseline_fitter.snip(j, max_half_window=max_half_window, decreasing=True, smooth_half_window=0)[0]
            
            # Save the baseline for other purpuses
            self.baseline.append(baseline_fitter.snip(j, max_half_window=max_half_window, decreasing=True, smooth_half_window=0)[0])
            
            # Set data below 0 to be 0 if exsisting
            if min(self.I[i]) < 0:
                self.I[i] = self.I[i].clip(min=0)
                
        return self.logger, self.I, self.baseline

    #----------------------------------------------------------------------------------------#
    #                                          Filter                                        #
    #----------------------------------------------------------------------------------------#
    def Filter(self,Filter,strength,order: int=2):
        """ Use a Savatzky-Golay or Whittaker-Eilers filter for the XRD/TS data.
        
        Arguments
        ----------
        Filter:     Type of Filter to use (lFilter or savgol)
        strength:   Strength with what the filter is applied 
        order:      Order of polinominal magnitude 
        """
        
        # Logging the event
        self.logger['Data was filtered'] = {   'Filtertype': Filter,
                                                'Used window lenght': strength,
                                                'Order of polinominal magnitude':order}
        
        # Loop over every I(q) from each scan
        for i,j in enumerate(self.I):
            
            # Applying a filter if set
            if Filter == "savgol":
                    self.I[i] = savgol_filter(j,strength,order)
            elif Filter == "whittaker":
                self.I[i] = whittaker_smooth(j,strength,d=order)
            
        return self.logger, self.I



    #----------------------------------------------------------------------------------------#
    #                                      Interpolation                                     #
    #----------------------------------------------------------------------------------------#
    def smooth_interpolation(self, output_multiplier):
        """ Interpolate datapoints in a smooth fashion.
        
        Arguments
        ----------
        output_multiplier:  number by which the data points are increased
        """
        
        # Logging the event
        self.logger['Data was interpolated'] = {'Points were multiplied by': output_multiplier,}
        
        # Loop over every I(q) from each scan
        for i,j in enumerate(self.I):
            
            # Determine the original number of points in the array
            n_points = len(j)
            
            # Create a new array of x-values for the output array
            x_new = np.linspace(self.q[0], self.q[-1], n_points*output_multiplier)
            
            # Use cubic spline interpolation to interpolate the data
            self.I[i] = np.interp(x_new, self.q, j, period=n_points)
            
        # Update the q array
        self.q = x_new
        return self.logger, self.q, self.I

    #----------------------------------------------------------------------------------------#
    #                              generate the PDF with PDFgetX3                            #
    #----------------------------------------------------------------------------------------#
    def generatePDF(self, dataPath, bgPath, outputPath, Composition, Qmaxinst, Qmax, Qmin, Rmax, Rpoly, scatter_supression: float=0.0):
        """ Using PDFgetX3 as a subprocess we can integrate the programm into our PDF generation.
            !!! Caution in some consoles this gives an error (in Visual Studio Code I can not run this, but in the powershell or bash I can) !!!

        Args:
            dataPath (str): Path to .dat files of the meas signal.
            bgPath (str):  Path to .dat files of the background signal, if already subtracted set to None.
            outputPath (str): Path to the output directory.
            Composition (_type_): Molecular formula of the material of interest.
           Qmin (float): Minuimum Q to include in the PDF generation. Defaults to 0.1.
            Qmax (float): Maximum Q to include in the PDF generation. Defaults to 30.
            Rpoly (float): Polynomal degree of data correction applied (best case: rpoly < Qmin). Defaults to 0.9.
            Rstep (float): Step with of the PDF (Smaller steps result in longer calculation times). Defaults to 0.01.
            Rmin (float): Minimum r the PDF is calculated from. Defaults to 0.5.
            Rmax (float): Maximum r the PDF is calculated to. Defaults to 30.
            scatter_supression (float, optional): Reduce the BG factor by a constant ammount to counter outliers. Defaults to 0.0.

        References:
            PDFgetX3:
                Code available from https://www.diffpy.org/products/pdfgetx.html
                Related Paper:  P. Juhás and T. Davis, C. L. Farrow, S. J. L. Billinge,https://journals.iucr.org/j/issues/2013/02/00/nb5046/index.html
                                J. Appl. Crystallogr. 46, 560-566 (2013)
    
        Returns:
            self.looger: updates the log file
        """
        
        # Logging the event
        self.logger['PDF was generated'] = {'Path to the data files': dataPath,
                                            'Path to the background files': bgPath,
                                            'Path to output directory': outputPath,
                                            'Compostion': Composition,
                                            'Qmaxinst': Qmaxinst,
                                            'Qmax':Qmax,
                                            'Qmin': Qmin,
                                            'Rmax': Rmax,
                                            'Rpoly': Rpoly}
        
        # Delete older folder contains (thus different averaging doesn´t effect storing new data in the same directory)
        if os.path.exists(outputPath):
            del_folder_content(outputPath)
        
        # Create directory if not exsiting
        ensure_dir(outputPath)
        
        
        i = 0
        bg_file_list = []
        
        # Get the names of the BG files in the given directory
        if bgPath != None:
            for bg_file in sorted(pathlib.Path(bgPath).iterdir()):
                bg_file_list.append(bg_file) 
        
        # Get the names of the files in the given directory
        for file in sorted(pathlib.Path(dataPath).iterdir()):
            
            # If BG files are given we can subtract the now:
            if bgPath != None:
                
                # We use a try function to skip data which has no BG files (can ouccure at the end if one or two scans don´t have a fitting background through averaging and rounding)
                try:
                    if  'meas_type1_uncorrected_Dat' in dataPath: 
                        process = subprocess.Popen(['pdfgetx3',str(file), '--force=true', '--format=QA', '--mode=xray', '--composition='+str(Composition), '--qmaxinst='+str(Qmaxinst), '--qmin='+str(Qmin), '--qmax='+str(Qmax), '--rpoly='+str(Rpoly), '--rmax='+str(Rmax),'-t=gr', '-b='+str(bg_file_list[int(i+self.bg_offset)]), '--bgscale='+str(self.factor_list[i]-scatter_supression), '-o='+str(outputPath)+'/@b'], stdout=subprocess.PIPE, text=True)
                    if  'meas_type2_corrected_prior_Dat' in dataPath: 
                        process = subprocess.Popen(['pdfgetx3',str(file), '--force=true', '--format=QA', '--mode=xray', '--composition='+str(Composition), '--qmaxinst='+str(Qmaxinst), '--qmin='+str(Qmin), '--qmax='+str(Qmax), '--rpoly='+str(Rpoly), '--rmax='+str(Rmax),'-t=gr', '-b='+str(bg_file_list[int(i+self.bg_offset)]), '--bgscale='+str(self.factor_list[i]), '-o='+str(outputPath)+'/@b'], stdout=subprocess.PIPE, text=True)
                    else:
                        process = subprocess.Popen(['pdfgetx3',str(file), '--force=true', '--format=QA', '--mode=xray', '--composition='+str(Composition), '--qmaxinst='+str(Qmaxinst), '--qmin='+str(Qmin), '--qmax='+str(Qmax), '--rpoly='+str(Rpoly), '--rmax='+str(Rmax),'-t=gr', '-b='+str(bg_file_list[int(i)]), '--bgscale='+str(self.factor_list[i]), '-o='+str(outputPath)+'/@b'], stdout=subprocess.PIPE, text=True)
                except:
                    # Give the user feedback about some missing BG files at the end (can be ignord normaly)
                    print('unoptimized BG')
            
            # If no BG is given we generate the PDF here
            else:   
                process = subprocess.Popen(['pdfgetx3',str(file), '--force=true', '--format=QA', '--mode=xray', '--composition='+str(Composition), '--qmaxinst='+str(Qmaxinst), '--qmin='+str(Qmin), '--qmax='+str(Qmax), '--rpoly='+str(Rpoly), '--rmax='+str(Rmax),'-t=gr', '-o='+str(outputPath)+'/@b'], stdout=subprocess.PIPE, text=True)
            
            # Get the consol outputs from PDFgetX3 and print them to the current consol
            while True:
                output = process.stdout.readline()
                if output:
                    print(output.strip())
                result = process.poll()
                if result is not None:
                    break
            i = i + 1
            
        return self.logger


    #----------------------------------------------------------------------------------------#
    #                              generate the PDF with PDFgetX3                            #
    #----------------------------------------------------------------------------------------#
    def generateGlasyPDF(self, dataPath, outputPath, Qmaxinst, Qmax, Qmin):
        """ 
        
        Arguments
        ----------
        dataPath:       path to .dat files of the meas signal    
        bgPath:         path to .dat files of the background signal, if already subtracted set to None
        outoutPath:     path to the output directory 
        Composition:    composition of the material of interest
        Qmaxinst:       maximum of q which correlates with background 
        Qmax:           maximum of q to be used for the Fourier Transform
        Qmin:           minimum of q to be used for the Fourier Transform
        Rmax:           maximum of r after the Fourier Transform
        Rpoly:          background correction factor from PDFgetX3
        """
        
        # Logging the event
        self.logger['Glasy PDF was generated'] = {'Path to the data files': dataPath,
                                            'Path to output directory': outputPath,
                                            'Compostion': 'SiO2',
                                            'Qmaxinst': Qmaxinst,
                                            'Qmax':Qmax,
                                            'Qmin': Qmin,
                                            'Rmin': 1.5,
                                            'Rmax': 1.7,
                                            'Rstep': 0.001,
                                            }
        
        self.glas_Gr = True
        if os.path.exists(outputPath):
            del_folder_content(outputPath)
            
        # Create directory if not exsiting
        ensure_dir(outputPath)
        
        
        i = 0
        for file in sorted(pathlib.Path(dataPath).iterdir()):
            process = subprocess.Popen(['pdfgetx3',str(file), '--force=true', '--format=QA', '--mode=xray', '--composition=SiO2', '--qmaxinst='+str(Qmaxinst), '--qmin='+str(Qmin), '--qmax='+str(Qmax), '--rpoly=0.9', '--rstep=0.001', '--rmin=1.5', '--rmax=1.7','-t=gr', '--bgscale=0','-o='+str(outputPath)+'/@b'], stdout=subprocess.PIPE, text=True)
            while True:
                output = process.stdout.readline()
                if output:
                    print(output.strip())
                result = process.poll()
                if result is not None:
                    break
            i = i + 1
        return self.logger, self.glas_Gr

    def generateMaximizationPDF(self, file, bg_file, factor, Rstep, Rmin, Rmax, outputPath, do_simple_correction: bool=False, do_iv_correction: bool=False):
        """ Similar to the generatePDF function it creates the PDF, 
            however this function is modified to rapidly maximize the PDF in the wanted area. 
            The function is called by the monte_carlo_PDF_maximization function and comes with some correction therms 
            for oversubtracted Y-Axis flipped PDFs.
        
        Arguments
        ----------
        dataPath:       path to .dat files of the meas signal    
        bgPath:         path to .dat files of the background signal, if already subtracted set to None
        outoutPath:     path to the output directory 
        Composition:    composition of the material of interest
        Qmaxinst:       maximum of q which correlates with background 
        Qmax:           maximum of q to be used for the Fourier Transform
        Qmin:           minimum of q to be used for the Fourier Transform
        Rmax:           maximum of r after the Fourier Transform
        Rpoly:          background correction factor from PDFgetX3
        """
        
        # Logging the event (needs to be adopded for iterative purposes)
        # self.logger['PDF was generated'] = {'Path to the data files': dataPath,
        #                                     'Path to the background files': bgPath,
        #                                     'Path to output directory': outputPath,
        #                                     'Compostion': Composition,
        #                                     'Qmaxinst': Qmaxinst,
        #                                     'Qmax':Qmax,
        #                                     'Qmin': Qmin,
        #                                     'Rmax': Rmax,
        #                                     'Rpoly': Rpoly}
        
        # Create directory if not exsiting
        ensure_dir(outputPath)
        
        # Get the name of the current file
        filename = os.path.basename(file).split("/")[-1].split(".")[0] + ".gr"

        # Generate the PDF with PDFgetX3
        process = subprocess.Popen(['pdfgetx3',str(file), '--force=true', '--format=QA', '--mode=xray', '--composition='+str(self.Composition), '--qmaxinst='+str(self.Qmaxinst), '--qmin='+str(self.Qmin), '--qmax='+str(self.Qmax), '--rpoly='+str(self.Rpoly), '--rstep='+str(Rstep), '--rmin='+str(Rmin), '--rmax='+str(Rmax),'-t=gr', '-b='+str(bg_file), '--bgscale='+str(factor), '-o='+str(outputPath)+'/@b'], stdout=subprocess.PIPE, text=True)
            
        # Get the consol outputs from PDFgetX3 and print them to the current consol
        while True:
            output = process.stdout.readline()
            if output:
                print(output.strip())
            result = process.poll()
            if result is not None:
                break
            
        # Reading the created .gr file to correct for oversubtraction if needed.
        self.r, self.BGmaximized_Gr = np.loadtxt(str(outputPath)+"/"+str(filename),skiprows=27).T
        
        # The simple correction works as long as the precursor is identifiable and PDF has a over net negative integral.
        if do_simple_correction == True:
            
            q, I_meas = np.loadtxt(file,skiprows=1).T
            q_bg, I_bg = np.loadtxt(bg_file,skiprows=1).T
            I_meas = I_meas[:self.clostest_Qmaxinst]
            
            
            from scipy.integrate import simpson
            if np.trapz(I_meas - self.factor*I_bg, dx=1e-10) < self.alltime_low_area:
                self.alltime_low_area = np.trapz(I_meas - self.factor*I_bg, dx=1e-10)
            if simpson(I_meas - self.factor*I_bg, dx=1e-10) < self.alltime_low_area2:
                self.alltime_low_area2 = simpson(I_meas - self.factor*I_bg, dx=1e-10)
            print(self.alltime_low)
            print(self.alltime_low_area)
            print(self.alltime_low_area2)
            
            if np.trapz(I_meas - self.factor*I_bg, dx=1e-10) < 0:
                self.flip = True 
                
            # if np.trapz(self.current_I - self.factor*self.I_bg) < 0:
            #     self.flip = True  
            
            # if np.trapz(self.BGmaximized_Gr) > 0:
            #     self.flip = True         

        # Experimantal function, that tries to get the phase shift of the FT noise.
        if do_iv_correction == True:
            
            # self.lowpass_BGmaximized_Gr = Filter(self.BGmaximized_Gr,"whittaker",1e3)[20:-20]     
            
            self.inversion_correction(True) # sets self.flip 
            
   
            
        # Inverting the current PDF and updating the .gr file.     
        if self.flip == True:
            print("\n\n G(r) inverted to -G(r) \n\n")
            overwrite_file(str(outputPath)+"/"+str(filename),self.r, -self.BGmaximized_Gr)
            self.BGmaximized_Gr = -self.BGmaximized_Gr  
            
        return max(self.BGmaximized_Gr) #self.logger,


    def monte_carlo_PDF_maximization(self, file, bg_file, outputPath, conversion_step_width: float=0.00001, Qlimits: list=[1.8,2.5],
                                     progress_critierium: int=20, reducing_steps: int=5, starting_step_width: float=0.1, scan_number: float=None):
        """ Calculation of the BG factor by maximisation of the PDF in the expected material r-range using a monte carlo style approch. 

        Args:
            file (str): Path to the .dat file used for the probed PDF generation.
            bg_file (str): Path to the .dat file of the BG used for the probed PDF generation
            outputPath (str): Path to the directory the probed PDFs should be stored in. 
            conversion_step_width (float, optional): Range around the highest value, at which conversion is sattisfied. Defaults to 0.00001.
            Qlimits (list): List containing the Q-range for the PDF maximization [lowest Q, higherst Q].Defaults to [2,4].
            progress_critierium (int, optional): Number of total steps after which the step width is reduced
                                                (if a monte carlo guess is very good in the beginning 
                                                it takes much longer for it to guess improved values). Defaults to 20.
            reducing_steps (int, optional): Number of improved steps after which the step width is reduced. Defaults to 5.
            starting_step_width (float, optional): Starting scattering range allowed for the random values generated around the current best value. Defaults to 0.1.
            scan_number (float): Number of the cuirrent scan in progress.

        Returns:
            best value (float): BG factor which yielded the highest G(r).
            score (float): highest G(r) probed.
        """
        
        # We use the fast processed factor as a initial guess, alternitavly one can do #random.uniform(min_value, max_value)
        best_value = self.factor 
        
        # Reset the correction varible for each scan
        self.flip = False
        
        # Set score value for initial guess
        best_score = self.generateMaximizationPDF(file,bg_file, best_value,0.1,Qlimits[0],Qlimits[1],outputPath) # ----- self.min_mc_r, self.max_mc_r -------
        
        score_steps = 0
        q = 0
        step_width = starting_step_width
        
        # Iterate until converging (happens befor 1000)
        for _ in range(1000):
            
            # Generate a random facor within the given range
            x = random.uniform(best_value - step_width, best_value + step_width)
            
            # calculate the score for it
            max_y = self.generateMaximizationPDF(file,bg_file, x, self.Rstep,Qlimits[0],Qlimits[1],outputPath)
            
            # self.min_y = 
            # self.max_y =
            
            # Test is the score is better the current best value
            if max_y > best_score:
                
                score_steps = score_steps +1

                # Update best score and best value
                best_score = max_y
                best_value = x
            
            # If best score is broken reducing_steps times reduce the step_width arount the best value
            if score_steps > reducing_steps or q > progress_critierium:
                step_width = step_width*0.1
                q = 0
                score_steps = 0
                
            # If step_with is sufgficient convergion is reached 
            if step_width < conversion_step_width:
                print("\n---- Convergence achieved ----")
                print("Best BG factor is: " + str( best_value)+ "  ,with a G(r) of: " + str(best_score) + "  ,of current scan: " + str(int(scan_number+1)) + "/" + str(int(len(self.I)))+ "\n")
                break
            else:
                print("BG factor: " + str(x)+ "  ,G(r): " + str(max_y) + "  ,scan: " + str(int(scan_number+1)) + "/" + str(int(len(self.I))))
                
            q = q + 1

        
        # Experimantal inversion correction
        #self.generateMaximizationPDF(file,bg_file, best_value, self.Rstep,35,40,outputPath, do_iv_correction=True)
        # self.generateMaximizationPDF(file,bg_file, best_value, self.Rstep*10,self.Rmax+20,self.Rmax+50,outputPath, do_iv_correction=True)
        
        # Create full PDF with good r resolution from best value
        self.generateMaximizationPDF(file,bg_file, best_value, self.Rstep,self.Rmin,self.Rmax,outputPath)
        return best_value, best_score
    
    def monte_carlo_PDF_maximization_with_glas_removal(self, file, bg_file, glas_file, outputPath,  starting_step_width: float=0.1, 
                                                        starting_step_width_glas=0.001, conversion_step_width: float=0.00001, progress_critierium: int=20, 
                                                        reducing_steps: int=5,  scan_number: float=None, Qlimits: list=[2,4]):
        """ Calculation of the BG factor by maximisation of the PDF in the expected material r-range using a monte carlo style approch. 

        Args:
            file (str): Path to the .dat file used for the probed PDF generation.
            bg_file (str): Path to the .dat file of the BG used for the probed PDF generation
            outputPath (str): Path to the directory the probed PDFs should be stored in. 
            conversion_step_width (float, optional): Range around the highest value, at which conversion is sattisfied. Defaults to 0.00001.
            progress_critierium (int, optional): Number of total steps after which the step width is reduced
                                                (if a monte carlo guess is very good in the beginning 
                                                it takes much longer for it to guess improved values). Defaults to 20.
            reducing_steps (int, optional): Number of improved steps after which the step width is reduced. Defaults to 5.
            starting_step_width (float, optional): Starting scattering range allowed for the random values generated around the current best value. Defaults to 0.1.
            starting_step_width_glas (float, optional): Starting scattering range allowed for the random values generated around the current best value 
                                                        for the glas contribution. Depending on the integration time and averaging from the empty glas. Defaults to 0.001.
            scan_number (float): Number of the cuirrent scan in progress.
            Qlimits (list): List containing the Q-range for the PDF maximization [lowest Q, higherst Q].Defaults to [2,4].
            
        Returns:
            best value (float): BG factor which yielded the highest G(r).
            score (float): highest G(r) probed.
        """
        
        def alter_bg_file(glas_factor):
            """ Takes a BG file and removes / adds glas contribution to it if needed

            Args:
                glas_factor (float): Intensity of the glas contribution

            Returns:
                new_bg_filepath (str): Path to the BG with glas contribution subtracted / added
            """
            # Reading the q and I(q) from the BG file and from the glas contribution file
            bg_q, bg_I = np.loadtxt(bg_file,skiprows=1).T
            glas_q, glas_I = np.loadtxt(glas_file,skiprows=1).T
            
            # I(q) from the glas is subtracted by the amount given from I(q) from the BG
            bg_I = bg_I - glas_factor * glas_I[:len(bg_I)]

            # Getting the number of the scan from the BG path, 
            filenumber = bg_file.split("_")[-1].split(".")[0]
            new_bg_filepath = rootdir + '/../../processed/' +self.sampleName +'/Dat_files/bg_glas_monte_Dat/' 
            
            # Data is stored as a new BG file for the PDF generation with PDFgetX3
            outputDatFile(new_bg_filepath,bg_q,[bg_I],filenumber,'background_glas_rm_'+self.sampleName, clear_folder=True)
            
            # The file path of the altered data is returned
            return new_bg_filepath + 'background_glas_rm_'+self.sampleName + "_" + filenumber + ".txt"
 
        
        # We use the fast processed factor as a initial guess, alternitavly one can do #random.uniform(min_value, max_value)
        best_value = self.factor 
        best_glas_value = 0
        
        # Reset the correction varible for each scan
        self.flip = False
        
        # Set score value for initial guess
        best_score = self.generateMaximizationPDF(file,bg_file, best_value,0.1,Qlimits[0],Qlimits[1],outputPath) # ----- self.min_mc_r, self.max_mc_r -------
        
        score_steps = 0
        q = 0
        step_width = starting_step_width
        step_width_glas = starting_step_width_glas
        
        # Iterate until converging (happens befor 1000)
        for _ in range(1000):
            
            if q % 2 == 0:
                x_glas = random.uniform(best_glas_value - step_width_glas, best_glas_value + step_width_glas)
                bg_file = alter_bg_file(x_glas)    
                y = self.generateMaximizationPDF(file,bg_file, x, 0.1,Qlimits[0],Qlimits[1],outputPath)
            else:
                # Generate a random facor within the given range
                x = random.uniform(best_value - step_width, best_value + step_width)
                
                # calculate the score for it
                y = self.generateMaximizationPDF(file,bg_file, x, 0.1,Qlimits[0],Qlimits[1],outputPath)
            
            # Test is the score is better the current best value
            if y > best_score:
                
                score_steps = score_steps +1

                # Update best score and best value
                best_score = y
                best_value = x
                best_glas_value = x_glas
            
            # If best score is broken reducing_steps times reduce the step_width arount the best value
            if score_steps > reducing_steps or q > progress_critierium:
                step_width = step_width*0.1
                step_width_glas = step_width_glas*0.1
                q = 0
                score_steps = 0
                
            # If step_with is sufgficient convergion is reached 
            if step_width < conversion_step_width:
                print("\n---- Convergence achieved ----")
                print("Best BG factor is: " + str( best_value)+ "  ,with a G(r) of: " + str(best_score) + "  ,of current scan: " + str(int(scan_number+1)) + "/" + str(int(len(self.I)))+ "\n")
                break
            else:
                print("BG factor: " + str(x)+ "  ,G(r): " + str(y) + "  ,scan: " + str(int(scan_number+1)) + "/" + str(int(len(self.I))))
                
            q = q + 1        
        
        # Create full PDF with good r resolution from best value
        self.generateMaximizationPDF(file,bg_file, best_value, self.Rstep,self.Rmin,self.Rmax,outputPath)
        return best_value, best_score
    
    
    # def inversion_correction(self,kind: bool):
    #     """ Flipps oversubtracted data by comparing oszilations at high r.

    #     Args:
    #         minR (float): Minimum r of the r-range of only Fopurier oszilations at high r.
    #         maxR (float): Maximum r of the r-range of only Fopurier oszilations at high r.
    #         kind (bool): Bool to choose if negativ phase offset is True or if the phase offset is positiv (False).

    #     Returns:
    #         G(r): With flipped oversubtracted scans.
    #     """
    #     # Using the sine fit function on G(r)
    #     # offset, phase = fit_sin(self.r[20:-20],self.lowpass_BGmaximized_Gr)
    #     # print(offset,phase)
        
    #     offset = fit_modulated_wave(self.r,self.BGmaximized_Gr)
    #     print(offset)
        
    #     # Set the flip varible to True to change G(r) to -G(r)
    #     if (offset < 0) != kind:
    #         self.flip = True
                