from datetime import datetime
import numpy as np
import os

class Import_PDF_Plot:
    """ Class to import PDFs from .gr files and temperature from .metadata
    """
    
    def import_data(self):
        """Imports r and G(r) from PDF .gr files.

        Returns:
            self.r: r of PDF data (X-Axis)
            self.Gr: G(r) of PDF data (Y-Axis)
        """
        
        # Make a list of the .gr data in directroy
        list_files = [f for f in os.listdir(self.path) if f.endswith('.gr') if not "dark" in f]
        
        # Sort them by number
        list_files.sort()

        # Loop over every scan 
        data = []
        for frame in list_files:
            counter = 0
            
            # Read data from file line by line
            with open(self.path+frame, 'r') as r: 
                lines = r.readlines()
                
                # Detect how many lines are headerlines (can differ some times)
                for i, line in enumerate(lines): 
                    
                    # Typical key from PDFgetX3 data
                    if str(line) == "#L r(Å)  G(Å$^{-2}$)\n" or str(line) == "#L r(Å)  G(Å$^{-2}$)" or str(line) == "#L r(Ã…)  G(Ã…$^{-2}$)\n" or str(line) == "#L r(Ã…)  G(Ã…$^{-2}$)": 
                        # Difine number of header lines
                        skip_lines = i + 1
            
                # Append the rest      
                data.append(np.genfromtxt(self.path+frame, skip_header=skip_lines))
        
        # Bring into good data format
        data = np.array(data)
        np.shape(data)

        # Extract arrays and return
        self.r = data[0,:,0]
        self.G = data[:,:,1] 
        return self.r, self.G
        
    #----------------------------------------------------------------------------------------#
    #                                get temp and time infos                                 #
    #----------------------------------------------------------------------------------------#
    def getTempTime(self,path,counter_PDF: dict):
        """Collects and averages the time and temperature points from the .metadata files.

        Args:
            path (str): Path to the metadata directory.
            counter_PDF (dict): names the comment on which time and temp follow in the metadata.

        Returns:
            self: Forwards updated self varibles to the main class.
        """
        
        # Define some arrays for the time and temperature data
        start_time = 0
        self.time_array = [] 
        self.time_average_array = []
        self.temp_array = []
        self.temp_average_array = []
        count = 1
 
        # Reject darks if wanted
        if self.without_dark == True:
            list_files = [f for f in os.listdir(path) if f.endswith('.metadata') if not "dark" in f]
        else:
            list_files = [f for f in os.listdir(path) if f.endswith('.metadata')]
        
        # Sort by number
        list_files.sort(key=str.lower)
        
        # Loop over all scans
        for file in list_files:
            file = path + file  
            f = open(file, "r")
            
            #Read line by line
            for line in f:
                line = line.rstrip()  # remove '\n' at end of line
                
                #Get time lines and set date string into seconds 
                if  counter_PDF['time'] in line:
                    
                    # Define start time as 0 seconds therefore this is the offset for all the other date strings
                    if start_time == 0:
                        start_time = datetime.strptime(line[int(len(counter_PDF['time'])):int(len(counter_PDF['time'])+10)] + " " + line[int(len(counter_PDF['time'])+13):], "%Y.%m.%d %H:%M:%S.%f")
                        self.time_array.append(0+self.timeOfset)
                    
                    else:
                        # Subtract start time from rest
                        delta_t = datetime.strptime(line[int(len(counter_PDF['time'])):int(len(counter_PDF['time'])+10)] + " " + line[int(len(counter_PDF['time'])+13):], "%Y.%m.%d %H:%M:%S.%f") - start_time
                        delta_t = delta_t.total_seconds() / 60
                        self.time_array.append(delta_t+self.timeOfset)
                        
                        # Set times for averaged data
                        if count % self.average == 0:
                            self.time_average_array.append(delta_t+self.timeOfset)
                            
                # Get temperature from the comment string
                if counter_PDF['temperature'] in line:
                    self.temp_array.append(float(line[int(len(counter_PDF['temperature'])):int(len(counter_PDF['temperature'])+8)]))
                    
                    # Set temperature for averaged data
                    if count % self.average == 0:
                            self.temp_average_array.append(float(line[int(len(counter_PDF['temperature'])):int(len(counter_PDF['temperature'])+8)]))
            
            f.close()
            count = count + 1
            
        # Crop unwanted pre or post scans
        self.time_average_array = self.time_average_array[self.startScan:self.endScan]
        self.temp_average_array = self.temp_average_array[self.startScan:self.endScan]
        return self.time_array, self.temp_array, self.temp_average_array, self.temp_average_array

    def reverse_PDF(self,style: str, data):
        """ Flipp the PDF if oversubtration occurs.

        Args:
            style (str):    PDF data can be reversed by "list" of every PDF index, 
                            "until" a certain index, "from a certain index", or "between" certain indices  
            data (int,list): list for style "list, float for "until" and "from", and list with size 2 for "between"

        Raises:
            AttributeError: Format errors for data

        Returns:
            list: Updated matrix of G
        """
        
        # Inverts G(r) for scans by their indizes put into a list
        if style == "list":
            try:
                for i in data:
                    self.G[i] = - self.G[i]
            except:
                raise AttributeError('Data not in format list')
        
        # Inverts G(r) for scans until a certain index
        if style == "until":
            try:
                for i,j in enumerate(self.G):
                    if i < data:
                        self.G[i] = - j
            except:
                raise AttributeError('Data not in format float')
            
        # Inverts G(r) for scans from a certain index
        if style == "from":
            try:
                for i,j in enumerate(self.G):
                    if i > data:
                        self.G[i] = - j
            except:
                raise AttributeError('Data not in format float')
        
        # Inverts G(r) for scans from a certain index to a certain index
        if style == "between":
            try:
                for i,j in enumerate(self.G):
                    if i > data[0]:
                        self.G[i] = - j
                    if i < data[1]:
                        self.G[i] = - j
            except:
                raise AttributeError('Data not in format list with size 2')
        
        return self.G  


    
class Import_PDF_Gen:
    """ Class to import întegrated data from .txt or .dat and time and temperature from .metadata files.
    """
    
    def import_data(self):
        """ Imports .chi or .dat for the PDF generation

        Raises:
            AttributeError: Raised, when no files in the given directory

        Returns:
            self.q: q from XRD/TS data (X-Axis)
            self.I: Intensity from XRD/TS data (Y-Axis)
        """
        
        # Make a list of the .gr data in directroy and reject darks if wanted
        if self.without_dark == True:
            list_files = [f for f in os.listdir(self.path) if f.endswith(self.ends_with) if not "dark" in f]
        else:
            list_files = [f for f in os.listdir(self.path) if f.endswith(self.ends_with)]
            
        # Sort them by number    
        list_files.sort(key=str.lower)
        
        # Raise error if directory is empty or files have wrong endings
        if list_files == []:
            raise AttributeError('No files with: ' + self.ends_with + '  found in directory: ' + self.path)
        
        # Calculate the number of files 
        if len(list_files) <= self.endScan:
            self.number_of_files = int(len(list_files)/self.number_of_steps)
        else:
            self.number_of_files = int(self.endScan/self.number_of_steps)

        # Setting some new lists
        self.q = []
        self.I = []
        
        # Loop over every scan 
        for i,j in enumerate(list_files):
            
            # Checking if scan is not unwanted pre or post data
            if i >= self.startScan and i <= self.endScan :
                
                # Depending on the integration setup dI might be given or not.
                # Here we account for both cases and load the file.
                try:
                    x, matrix, dI = np.loadtxt(self.path+j, skiprows=self.skiprows_of_files).T
                except:
                    x, matrix = np.loadtxt(self.path+j, skiprows=self.skiprows_of_files).T
                
                # Setup the data matrix (q should be same for all)
                self.q.append(x)
                self.I.append(matrix)
        
        # Numpy array from list for processing
        self.q = np.array(self.q[0])
        self.I = np.array(self.I)
        return self.q, self.I
    
    #----------------------------------------------------------------------------------------#
    #                                get temp and time infos                                 #
    #----------------------------------------------------------------------------------------#
    def getTempTime(self,path,counter_PDF: dict):
        """Collects and averages the time and temperature points from the .metadata files.

        Args:
            path (str): Path to the metadata directory.
            counter_PDF (dict): names the comment on which time and temp follow in the metadata.

        Returns:
            self: Forwards updated self varibles to the main class.
        """
        
        # Log the event
        self.logger['Temperature and Time loaded'] = {   'Path to .metadata': path,}
        
        # Define some arrays for the time and temperature data
        start_time = 0
        self.time_array = [] 
        self.time_average_array = []
        self.temp_array = []
        self.temp_average_array = []
        count = 1
        
        # Reject darks if wanted
        if self.without_dark == True:
            list_files = [f for f in os.listdir(path) if f.endswith('.metadata') if not "dark" in f]
        else:
            list_files = [f for f in os.listdir(path) if f.endswith('.metadata')]
            
        # Sort by number 
        list_files.sort(key=str.lower)
        
        # Loop over all scans
        for file in list_files:
            file = path + file  

            # Speed up loading to only needed data
            if count >= self.startScan and count <= self.endScan :
                if count % self.number_of_steps == 0:      
                    f = open(file, "r")
                    
                    #Read line by line
                    for line in f:
                        line = line.rstrip()  # remove '\n' at end of line
                        
                        #Get time lines and set date string into seconds 
                        if counter_PDF['time'] in line:
                            
                             # Define start time as 0 seconds therefore this is the offset for all the other date strings
                            if start_time == 0:
                                start_time = datetime.strptime(line[int(len(counter_PDF['time'])):int(len(counter_PDF['time'])+10)] + " " + line[int(len(counter_PDF['time'])+13):], "%Y.%m.%d %H:%M:%S.%f")
                                self.time_array.append(0+self.timeOfset)
                            
                            else:
                                # Subtract start time from rest
                                delta_t = datetime.strptime(line[int(len(counter_PDF['time'])):int(len(counter_PDF['time'])+10)] + " " + line[int(len(counter_PDF['time'])+13):], "%Y.%m.%d %H:%M:%S.%f") - start_time
                                delta_t = delta_t.total_seconds() / 60
                                self.time_array.append(delta_t+self.timeOfset)
                                
                                # Set times for averaged data
                                if count % self.average == 0:
                                    self.time_average_array.append(delta_t+self.timeOfset)
                                    
                        # Get temperature from the comment string
                        if counter_PDF['temperature'] in line:
                            self.temp_array.append(float(line[int(len(counter_PDF['temperature'])):int(len(counter_PDF['temperature'])+8)]))
                            
                            # Set temperature for averaged data
                            if count % self.average == 0:
                                    self.temp_average_array.append(float(line[int(len(counter_PDF['temperature'])):int(len(counter_PDF['temperature'])+8)]))
                    
                    f.close()
            count = count + 1
        return self.time_array, self.temp_array, self.time_average_array, self.temp_average_array, self.logger
    
    #----------------------------------------------------------------------------------------#
    #                                     Glass PDF loading                                  #
    #----------------------------------------------------------------------------------------#
    def GlassPDFload(self, input_dir):
        """Load glass file data to generate PDF of glass file.

        Args:
            input_dir (str): Path to input directory.

        Returns:
            self: updated varibles
        """
        # Log the event
        self.logger['BG scale determined by Glass Peak Intensity'] = {' ': 0,}
        
        # Varible, which keeps system up to date about t he event
        self.glassGr = True
        
        # Load PDF of the glas from .gr files ()
        list_files_glassy_Gr = [f for f in os.listdir(input_dir) if f.endswith('.gr')]
        
        #The list of data files is the sorted to confirm to the order 00001-99999 
        list_files_glassy_Gr.sort(key=str.lower)

        # Every single scan of the glas PDF will be stored in a matrix
        self.Gr = []
        for i,j in enumerate(list_files_glassy_Gr):
            r, Gr = np.loadtxt(input_dir + '/' + j,skiprows=26).T
            self.Gr.append(Gr)
        return self.Gr, self.glassGr   