import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib.widgets import CheckButtons
from matplotlib.ticker import LinearLocator, FormatStrFormatter  
import numpy as np
from scipy import optimize

import pathlib
rootdir = str(pathlib.Path(__file__).parent.resolve())
plt.style.use(rootdir + '/../settings/style.mplstyle')


from backbone.functions.utils import ensure_dir, find_nearest, bars_to_gaussian, exponential_radial_decay
import backbone.settings.Colormap as cmp


class Plottin_PDF:
    """Class containing different plotting styles perfected for in-situ PDF data
    """
    
    def inversion_correction(self,minR,maxR,kind: bool):
        """ Flipps oversubtracted data by comparing oszilations at high r.

        Args:
            minR (float): Minimum r of the r-range of only Fopurier oszilations at high r.
            maxR (float): Maximum r of the r-range of only Fopurier oszilations at high r.
            kind (bool): Bool to choose even or odd type Oszilations as normal.

        Returns:
            G(r): With flipped oversubtracted scans.
        """
        
        
        minR = find_nearest(self.r,minR)
        minR = int(np.where(self.r==minR)[0])
        maxR = find_nearest(self.r,maxR)
        maxR = int(np.where(self.r==maxR)[0])
        
        def fit_sin(tt, yy):
            """Fit sin to the input time sequence, and return fitting parameters "amp", "omega", "phase", "offset", "freq", "period" and "fitfunc"

            Args:
                tt (list, array): X-Axis of data the sin fit is applied to.
                yy (list, array): Y-Axis of data the sin fit is applied to.

            Returns:
                c (float): offset in y.
                p (float): phase coefficient.
            """
            tt = np.array(tt)
            yy = np.array(yy)
            ff = np.fft.fftfreq(len(tt), (tt[1]-tt[0]))   # assume uniform spacing
            Fyy = abs(np.fft.fft(yy))
            guess_freq = abs(ff[np.argmax(Fyy[1:])+1])   # excluding the zero frequency "peak", which is related to offset
            guess_amp = np.std(yy) * 2.**0.5
            guess_offset = np.mean(yy)
            guess = np.array([guess_amp, 2.*np.pi*guess_freq, 0., guess_offset])

            def sinfunc(t, A, w, p, c):  
                return A * np.sin(w*t + p) + c
            
            popt, pcov = optimize.curve_fit(sinfunc, tt, yy, p0=guess)
            A, w, p, c = popt
            f = w/(2.*np.pi)
            fitfunc = lambda t: A * np.sin(w*t + p) + c
            return c, p

        
        for i,j in enumerate(self.G): #insitu plot
            offset, phase = fit_sin(self.r,j)
            print(i,offset)
            if (offset < 0) != kind:
                self.G[i] = -j
                



    def plot1dvsTime(self, offsetTime = 0, title: str = None,savefig: str = None,outputPath=None, step_width=1, maxR=None, minR=None, getEndtime=False):
        """ Plot the in-situ data as a 1d plot overlaying each from first scan to last scan in order.

        Args:
            offsetTime (int, optional): Define time offset for the howl time array. Defaults to 0.
            title (str, optional): Plot title. Defaults to None.
            savefig (str, optional): Save the figure as a png or svg. Defaults to None.
            outputPath (_type_, optional): Path to the output directory for the saved figure. Defaults to None.
            step_width (int, optional): To reduce plotting time and or confusion from too many scans, scipping with a sertain width is possible. Defaults to 1.
            maxR (float, optional): Crop plot to a maximum r in Ang. Defaults to None.
            minR (float, optional): Crop plot to a minimum r in Ang. Defaults to None.
            getEndtime (bool, optional): Print out last time value to the consol. Defaults to False.
            
        """
        
        # Generates directory if not exsiting
        if outputPath != None:
            ensure_dir(outputPath)
            
        # Handel unnamed plots to get sample name as title
        if title == None:
            title = self.sampleName

        # Set up the plot
        fig, ax = plt.subplots()

        # Define maxima and minima for energy and intensity for ploting purpose
        self.max_x = max(self.r)
        self.min_x = min(self.r)
        
        # Get the PDF with highest G(r) for scaling the plot
        for i in self.G:
            if max(i) > self.max_all_int:
                self.max_all_int = max(i)
            if min(i) < self.min_all_int:
                self.min_all_int = min(i)

        # Set colormap
        colors = plt.cm.jet(np.linspace(0, 1, len(self.G)))
        
        # Loop over all scans
        for i,j in enumerate(self.G): #insitu plot
            
            # Skip scans which are not a multiple of the step width 
            if i % step_width == 0:

                # Make start and end scans appear thicker
                if len(self.G) % 2 ==0:
                    ramp = np.linspace(3,3/2,int(len(self.G)/2)) 
                    thickness = np.concatenate((ramp,ramp[::-1])) 
                else:
                    ramp = np.linspace(3,3/2,int(len(self.G)/2)) 
                    thickness = np.concatenate((np.concatenate((ramp,np.array([3/2]))),ramp[::-1])) 

                # If time starts when temperature is reached usefull to label data with pre stamp.
                if self.time_averaged_array[i] < offsetTime:
                    prelabel = "pre "
                else:
                    prelabel = ""

                # add Scan and legend
                ax.plot(self.r, j, linewidth=thickness[i], color=colors[i], zorder=-i, label=prelabel + str(int(self.temp_averaged_array[i])) + " °C", alpha=1) 
                plt.legend(prop={'size': 10}, ncol=3,loc='upper right').get_frame().set_linewidth(0.0)

        # Add colormap
        sm = plt.cm.ScalarMappable(cmap='jet', norm=plt.Normalize(vmin=int(self.time_averaged_array[step_width]), vmax=int(self.time_averaged_array[-1])))
        cb = plt.colorbar(sm)
        cb.set_label(label='time [min]')

        # User feedback (not realy nessicarry)
        if getEndtime == True:
            print(self.time_averaged_array[-1])
            
        
        # Set Axis and title names
        ax.set_title(title)    
        ax.set_xlabel("r [$\\rm{\AA}$]")
        ax.set_ylabel("G [$\\rm{\AA^{-2}}$]")

        # Crop data in X 
        if maxR != None and minR !=None:
            plt.xlim(minR,maxR)
        elif maxR != None and minR == None:
            plt.xlim(self.min_x,maxR)
        elif minR != None and maxR == None:
            plt.xlim(minR,self.max_x)
        elif maxR == None and minR == None:
            plt.xlim(self.min_x,self.max_x)
            
        plt.ylim()
        
        # Save plot
        if savefig == "png":
            plt.savefig(outputPath + "/1d_"+ self.sampleName +".png", dpi=600)
        if savefig == "svg":
            plt.savefig(outputPath + "/1d_"+ self.sampleName +".svg", dpi=600)
        plt.show()


    def plot1dvsTimeProjectedOntop(self, offsetTime = 0, title: str = None,savefig: str = None,outputPath=None, step_width=1, maxR=None, minR=None, getEndtime=False, intercept: float=0.3):
        """ Plot the in-situ data as a 1d plot with a fixed offset for each scan from first scan to last scan in order.

        Args:
            offsetTime (int, optional): Define time offset for the howl time array. Defaults to 0.
            title (str, optional): Plot title. Defaults to None.
            savefig (str, optional): Save the figure as a png or svg. Defaults to None.
            outputPath (_type_, optional): Path to the output directory for the saved figure. Defaults to None.
            step_width (int, optional): To reduce plotting time and or confusion from too many scans, scipping with a sertain width is possible. Defaults to 1.
            maxR (float, optional): Crop plot to a maximum r in Ang. Defaults to None.
            minR (float, optional): Crop plot to a minimum r in Ang. Defaults to None.
            getEndtime (bool, optional): Print out last time value to the consol. Defaults to False.
            intercept (float, optional): Offset choosen for each scan. Defaults to 0.3.
            
        """
        
        # Generates directory if not exsiting
        if outputPath != None:
            ensure_dir(outputPath)
            
        # Handel unnamed plots to get sample name as title
        if title == None:
            title = self.sampleName

        # Set up the plot
        fig, ax = plt.subplots()

        #define maxima and minima for energy and intensity for ploting purpose
        self.max_x = max(self.r)
        self.min_x = min(self.r)
        
        
        # Get the PDF with highest G(r) for scaling the plot
        for i in self.G:
            if max(i) > self.max_all_int:
                self.max_all_int = max(i)
            if min(i) < self.min_all_int:
                self.min_all_int = min(i)

        height = 0
        
        # Set colormap
        colors = plt.cm.jet(np.linspace(0, 1, len(self.G)))
        
        # Loop over all scans
        for i,j in enumerate(self.G): #insitu plot
            
            # Skip scans which are not a multiple of the step width 
            if i % step_width == 0:
                
                # Make start and end scans appear thicker
                if len(self.G) % 2 ==0:
                    ramp = np.linspace(3,3/2,int(len(self.G)/2)) 
                    thickness = np.concatenate((ramp,ramp[::-1])) 
                else:
                    ramp = np.linspace(3,3/2,int(len(self.G)/2)) 
                    thickness = np.concatenate((np.concatenate((ramp,np.array([3/2]))),ramp[::-1])) 
                
                # If time starts when temperature is reached usefull to label data with pre stamp.
                if self.time_averaged_array[i] < offsetTime:
                    prelabel = "pre "
                else:
                    prelabel = ""

                # add Scan and legend
                ax.plot(self.r, j+height, linewidth=thickness[i], color=colors[i], zorder=-i, label=prelabel + str(int(self.temp_averaged_array[i])) + " °C", alpha=1) 
                plt.legend(prop={'size': 10}, ncol=3,loc='upper right').get_frame().set_linewidth(0.0)
                height = height + max(j) + intercept*abs(min(j))
        
        # Add colormap
        sm = plt.cm.ScalarMappable(cmap='jet', norm=plt.Normalize(vmin=int(self.time_averaged_array[step_width]), vmax=int(self.time_averaged_array[-1])))
        cb = plt.colorbar(sm)
        cb.set_label(label='time [min]')
        
        # User feedback (not realy nessicarry)
        if getEndtime == True:
            print(self.time_averaged_array[-1])
        
        # Set Axis and title names
        ax.set_title(title)    
        ax.set_xlabel("r [$\\rm{\AA}$]")
        ax.set_ylabel("G [$\\rm{\AA^{-2}}$]")

        # Crop data in X 
        if maxR != None and minR !=None:
            plt.xlim(minR,maxR)
        elif maxR != None and minR == None:
            plt.xlim(self.min_x,maxR)
        elif minR != None and maxR == None:
            plt.xlim(minR,self.max_x)
        elif maxR == None and minR == None:
            plt.xlim(self.min_x,self.max_x)
            
        # Crop data in Y to highest Scan + 3
        plt.ylim(top=height+3)
        
        # Save plot
        if savefig == "png":
            plt.savefig(outputPath + "/1d_"+ self.sampleName +".png", dpi=600)
        if savefig == "svg":
            plt.savefig(outputPath + "/1d_"+ self.sampleName +".svg", dpi=600)
        plt.show()




    #----------------------------------------------------------------------------------------#
    #                             plot data in 2D with temp curve                            #
    #----------------------------------------------------------------------------------------#
    def plot2dTemp(self,vmin=None,vmax=None,savefig: str = None,outputPath=None, rect=None, minR = None, maxR = None):
        """ Plot a 2D map of time and r with G(r) indicated by a colormap. The Temperature curved is paced side by side with the 2D plot.

        Args:
            vmin (_type_, optional): Intensity the colormap should start. Defaults to None.
            vmax (_type_, optional): Intensity the colormap should end. Defaults to None.
            savefig (str, optional): Save the figure as a png or svg. Defaults to None.
            outputPath (_type_, optional): Path to the output directory for the saved figure. Defaults to None.
            rect (list, optional): Pints around which a red rectangle should b e drawn (tool for insets). Defaults to None.
            maxR (float, optional): Crop plot to a maximum r in Ang. Defaults to None.
            minR (float, optional): Crop plot to a minimum r in Ang. Defaults to None.
            
        """
        
        # Generates directory if not exsiting
        if outputPath != None:
            ensure_dir(outputPath)
            
        # Set some plotting settings
        newcmp = cmp.cm2d()

        # Set up the plot
        fig, (ax1, ax2) = plt.subplots(1, 2, gridspec_kw={'width_ratios': [1, 4]},figsize=(self.figratio2D[0], self.figratio2D[1]),sharey='all')
        plt.subplots_adjust(wspace = 0.075 )

        # Crop the time and temperature to the wanted leght
        if self.reload == None:
            time = self.time_averaged_array[self.startScan:self.endScan]
            temp = self.temp_averaged_array[self.startScan:self.endScan]
        else:
            time = self.time_averaged_array
            temp= self.temp_averaged_array

        # Plot time vs temperature curve
        ax1.plot(temp,time, color="red")
        ax1.set(xlabel = 'T [°C]', ylabel="time [min]",ylim = (min(time),max(time)), xlim=(0,200))
        ax1.tick_params(axis='x',which="major", direction="out")
        
        # Here we assume temperatures between 100 and 200 °C to be the norm
        ax1.set_xticks([0,100,200])

        # Set x, y and z for the 2D colormap 
        x = self.r
        # Crop the time to the wanted leghtto the wanted leght
        if self.reload == None:
            y = self.time_averaged_array[self.startScan:self.endScan]
        else:
            y = self.time_averaged_array
        z = self.G

        # Plot the 2D colormap 
        imap = ax2.imshow(z, origin = 'lower', aspect = 'auto', 
                extent = [min(x), max(x), min(y), max(y)], vmin = vmin, vmax = vmax, 
                cmap=newcmp, interpolation='gaussian')

        # imap = ax2.contourf(x,y,z, origin = 'lower', levels=np.linspace(vmin, vmax, 1000), vmin = vmin, vmax = vmax, cmap=newcmp,extend='both')

        # Add a red rectange if wanted for inlays
        if rect != None:
            rect = patches.Rectangle((rect["minR"], rect["mintime"]), (rect["maxR"]-rect["minR"]), (rect["maxtime"]-rect["mintime"]), linewidth=3/2, edgecolor='r', facecolor='none',zorder=100)
            ax2.add_patch(rect)

        # Set X ticks
        ax2.set_xticks(np.arange(round(min(x),1),round(max(x)+1,2), 5))
        
        # Crop data in X 
        if maxR != None and minR !=None:
            ax2.set(xlim = (minR,maxR), xlabel = 'r [$\\rm{\AA}$]')
        elif minR != None and maxR == None:
            ax2.set(xlim = (minR,max(x)), xlabel = 'r [$\\rm{\AA}$]')
        elif maxR != None and minR == None:
            ax2.set(xlim = (min(x),maxR), xlabel = 'r [$\\rm{\AA}$]')
        elif maxR == None and minR == None:
            ax2.set(xlim = (min(x),max(x)), xlabel = 'r [$\\rm{\AA}$]')

        #Set ticks for both plots
        ax2.tick_params(axis="y",left=False,labelleft=False)
        ax2.tick_params(axis='both',which="major", direction='out')
        ax2.tick_params(axis='x',which="minor", width=1, length=3, labelsize=15, direction='out')
        
        # Set title name to be sample name
        ax2.set_title(self.sampleName)

        # Add colormap
        cb = plt.colorbar(imap, extendrect=True)
        cb.ax.tick_params(axis='both',which="both", width=1.7, length=3.5, labelsize=15, direction="in") 
        cb.ax.tick_params(axis='x',which="minor", width=1, length=2, labelsize=15,direction="in") 
        cb.set_label(label='G [$\\rm{\AA^{-2}}$]')

        # Save plot
        if savefig == "png":
            plt.savefig(outputPath + "/2d_"+ self.sampleName +".png", dpi=600)
        if savefig == "svg":
            plt.savefig(outputPath + "/2d_"+ self.sampleName +".svg", dpi=600)
        plt.show()        


    #----------------------------------------------------------------------------------------#
    #                                    plot data in 3D                                     #
    #----------------------------------------------------------------------------------------#
    def plot3d(self, vmin= None, vmax=None, elev=None, azim=None,savefig: str = None,outputPath=None):  
        """ Looks cool but represents the data not that well. 

        Args:
            vmin (_type_, optional): Intensity the colormap should start. Defaults to None.
            vmax (_type_, optional): Intensity the colormap should end. Defaults to None.
            elev (_type_, optional): Elevation angle 0 equals ground or 1d side perspective. Defaults to None.
            azim (_type_, optional): Azimutal angel -90 equals 1d front perspective. Defaults to None.
            savefig (str, optional): Save the figure as a png or svg. Defaults to None.
            outputPath (_type_, optional): Path to the output directory for the saved figure. Defaults to None.
            
        """
        
        # Generates directory if not exsiting
        if outputPath != None:
            ensure_dir(outputPath)
            
        
        # Set up the plot
        fig = plt.figure()
        ax = fig.gca(projection='3d')
        fig.subplots_adjust(top=1.1, bottom=-.1)

        # Set colormap
        newcmp = cmp.cm2d()
        
        # Plot the surface of the 3D model.
        x0, x1 = np.meshgrid(np.array(self.r), np.array(self.time_array))
        surf = ax.plot_surface(x0, x1, np.array(self.G), cmap=newcmp,
                            linewidth=-3, antialiased=False, rstride=1, cstride=1, alpha=1, vmin=vmin, vmax=vmax)

        # Customize the z axis.
        ax.zaxis.set_major_locator(LinearLocator(10))
        ax.zaxis.set_major_formatter(FormatStrFormatter('%.02f'))

        #ax.set_zticks([])
        ax.set_xlabel("Energy [keV]")
        ax.set_ylabel("time [min]")#"time [s]")
        ax.set_zlabel("Intensity [arb. units]")
        ax.view_init(elev=elev, azim=azim)

        # Add a color bar which maps values to colors.
        # fig.colorbar(surf, shrink=0.5, aspect=5)

        # Save plot
        if savefig == "png":
            plt.savefig(outputPath + "/3d_"+ self.sampleName +".png", dpi=600)
        if savefig == "svg":
            plt.savefig(outputPath + "/3d_"+ self.sampleName +".svg", dpi=600)
        plt.show()


class Plotting_XRD:
    """ Class to plot data from inverted space (XRD or TS data)
    """
    
    
    #----------------------------------------------------------------------------------------#
    #                             plot data in 2D with temp curve                            #
    #----------------------------------------------------------------------------------------#
    def plot2dTemp(self,vmin=None,vmax=None,savefig: str = None,outputPath=None, rect=None, maxTemp: float =200, minTemp: float =0):
        """ Plot a 2D map of time and q with I(q) indicated by a colormap. The Temperature curved is paced side by side with the 2D plot.

        Args:
            vmin (_type_, optional): Intensity the colormap should start. Defaults to None.
            vmax (_type_, optional): Intensity the colormap should end. Defaults to None.
            savefig (str, optional): Save the figure as a png or svg. Defaults to None.
            outputPath (_type_, optional): Path to the output directory for the saved figure. Defaults to None.
            rect (list, optional): Pints around which a red rectangle should b e drawn (tool for insets). Defaults to None.
            maxTemp (float, optional): Maximum temperature to scale the time-temperature plot. Defaults to 200.
            minTemp (float, optional): Minimum temperature to scale the time-temperature plot. Defaults to 0.
            

        """
        
        # Generates directory if not exsiting
        if outputPath != None:
            ensure_dir(outputPath)
        else:
            outputPath = self.path +'../processed/Plot_2D_Temp/' 
            ensure_dir(outputPath)
            
        # Get the scan with highest and lowest intensity
        for i in self.I:
            if max(i) > self.max_all_int:
                self.max_all_int = max(i)
            if min(i) < self.min_all_int:
                self.min_all_int = min(i)
        
        # Set maximum and minimum of the colormap to the lowest and highest values
        if vmin == None:
            vmin = self.max_all_int
        if vmax == None:
            vmax = self.min_all_int
            
        # Set colormap
        newcmp = cmp.cm2d()

        # Set up the plot
        fig, (ax1, ax2) = plt.subplots(1, 2, gridspec_kw={'width_ratios': [1, 4]},figsize=(self.figratio2D[0], self.figratio2D[1]),sharey='all')
        plt.subplots_adjust(wspace = 0.075 )


        # Set local varibles of time and temperature
        time = self.time_array
        temp = self.temp_array


        # Plot time vs temperature curve
        ax1.plot(temp,time, color="red")
        ax1.set(xlabel = 'T [°C]', ylabel="Time [min]",ylim = (min(time),max(time)), xlim=(minTemp,maxTemp)) 
        ax1.tick_params(axis='x',which="major", direction="out")

        # Set x, y and z for the 2D colormap 
        x = self.q
        y = self.time_array
        z = self.I

        # Plot the 2D colormap      
        imap = ax2.imshow(z, origin = 'lower', aspect = 'auto', 
                extent = [min(x), max(x), min(y), max(y)], vmin = vmin, vmax = vmax, 
                cmap=newcmp, interpolation='gaussian')

        # Add a red rectange if wanted for inlays
        if rect != None:
            rect = patches.Rectangle((rect["minR"], rect["mintime"]), (rect["maxR"]-rect["minR"]), (rect["maxtime"]-rect["mintime"]), linewidth=self.linewidth/2, edgecolor='r', facecolor='none',zorder=100)
            ax2.add_patch(rect)


        # Set ticks for both plots
        ax2.set(xlim = (min(x),max(x)), xlabel = 'q [$\\rm{\AA^{-1}}$]')
        ax2.tick_params(axis='x',which="major", direction="out")
        ax2.tick_params(axis="y",left=False,labelleft=False)
        ax2.set_title(self.sampleName)


        # Add colormap
        cb = plt.colorbar(imap)
        cb.set_label(label='I [arb. un.]')


        # Save plot
        if savefig == "png":
            plt.savefig(outputPath + "/2d_"+ self.sampleName +".png", dpi=600)
        if savefig == "svg":
            plt.savefig(outputPath + "/2d_"+ self.sampleName +".svg", dpi=600)
        plt.show()        


    #----------------------------------------------------------------------------------------#
    #                              Plot 1d Data vs Temp and Time                             #
    #----------------------------------------------------------------------------------------#
    def plot1dvsTime(self, offsetTime = 0, title: str = None,savefig: str = None,outputPath=None, step_width =1, maxQ=None, minQ=None, plot_temp: bool=False,min_index: int=0, max_index: int=-1):
        """
        
        Args:
            offsetTime (int, optional): Define time offset for the howl time array. Defaults to 0.
            title (str, optional): Plot title. Defaults to None.
            savefig (str, optional): Save the figure as a png or svg. Defaults to None.
            outputPath (_type_, optional): Path to the output directory for the saved figure. Defaults to None.
            step_width (int, optional): To reduce plotting time and or confusion from too many scans, scipping with a sertain width is possible. Defaults to 1.
            maxQ (float, optional): Crop the plot to a maximum of Q. Defaults to None.
            minQ (float, optional): Crop the plot to a minimum of Q. Defaults to None.
            plot_temp (bool, optional): Plot temperature color bar by side with the data. Defaults to False.
            min_index (int, optional): Smallest index for the color bar. Defaults to 0.
            max_index (int, optional): Largest index for the color bar. Defaults to -1.
            
        """
        
        # Generates directory if not exsiting
        if outputPath != None:
            ensure_dir(outputPath)
        else:
            ensure_dir(outputPath)
            outputPath = self.path +'/plot/' 
        if title == None:
            title = self.sampleName

        # Set up the plot
        plt.figure() #
        ax = plt.gca()     

        # Define maxima and minima for energy and intensity for ploting purpose
        self.max_x = max(self.q)
        self.min_x = min(self.q)
        for i in self.I:
            if max(i) > self.max_all_int:
                self.max_all_int = max(i)
            if min(i) < self.min_all_int:
                self.min_all_int = min(i)


        # Set colormap
        colormapRB = cmp.cmRB_Time(10000)
        colors = colormapRB(np.linspace(0,1,len(self.I[min_index:max_index])))
        
        # Loop over all scans
        for i,j in enumerate(self.I[min_index:max_index]): #insitu plot
            
            # Skip scans which are not a multiple of the step width 
            if i % step_width == 0:

                # Make start and end scans appear thicker# Make start and end scans appear thicker
                if len(self.I) % 2 ==0:
                    ramp = np.linspace(3,3,int(len(self.I)/2)) 
                    thickness = np.concatenate((ramp,ramp[::-1])) 
                else:
                    ramp = np.linspace(3,3/2,int(len(self.I)/2)) 
                    thickness = np.concatenate((np.concatenate((ramp,np.array([3/2]))),ramp[::-1])) 

                # If time starts when temperature is reached usefull to label data with pre stamp.
                if self.time_array[i] < offsetTime:
                    prelabel = "pre "
                else:
                    prelabel = ""
                
                # add Scan and legend if wanted
                ax.plot(self.q, j, linewidth=thickness[i], color=colors[i], zorder=len(self.I)-i, label=prelabel + str(int(self.temp_array[int(i*self.average)])) + " °C", alpha=1) 
                if plot_temp == True:
                    plt.legend(prop={'size': 10}, ncol=3,loc='upper right', handlelength=1.6).get_frame().set_linewidth(0.0)
       
        # Define colormap limits
        if max_index != -1:
            max_index_time = int(max_index*self.average)
        min_index_time = int(min_index*self.average)
        
        # Add colormap
        sm = plt.cm.ScalarMappable(cmap=colormapRB, norm=plt.Normalize(vmin=int(self.time_array[min_index_time]), vmax=int(self.time_array[max_index_time])))
        cb = plt.colorbar(sm)
        cb.set_label(label='Time [min]')


        # Set Axis and title names
        ax.set_title(title)    
        ax.set_xlabel("q [$\\rm{\AA^{-1}}$]")
        ax.set_ylabel("I [$\\rm{arb. un.}$]")

        # Crop data in X 
        if maxQ != None and minQ !=None:
            plt.xlim(minQ,maxQ)
        elif maxQ != None and minQ == None:
            plt.xlim(self.min_x,maxQ)
        elif minQ != None and maxQ == None:
            plt.xlim(minQ,self.max_x)
        elif maxQ == None and minQ == None:
            plt.xlim(self.min_x,self.max_x)
        
        # Crop data in Y
        plt.ylim(bottom=0)
        
         # Save plot
        if savefig == "png":
            plt.savefig(outputPath + "/1d_"+ self.sampleName +".png", dpi=600)
        if savefig == "svg":
            plt.savefig(outputPath + "/1d_"+ self.sampleName +".svg", dpi=600)
        plt.show()

    #----------------------------------------------------------------------------------------#
    #                                    Check the phases                                    #
    #----------------------------------------------------------------------------------------#
    def checkPhases(self,scan_number: int,references: dict, savefig: str = None, outputPath=None,  maxQ=None, minQ=None, energy: float=100):
        """ Plotting a choosen scan against a single or a set of references.

        Args:
            scan_number (int): Scan, at which the phase is probed.
            references (dict): References used to check the phase. Dict style, which can be put to reference sets of certain material classes.
            savefig (str, optional): Save the figure as a png or svg. Defaults to None.
            outputPath (_type_, optional): Path to the output directory for the saved figure. Defaults to None.
            maxQ (float, optional): Crop the plot to a maximum Q. Defaults to None.
            minQ (float, optional): Crop the plot to a minimum Q. Defaults to None.
            energy (float, optional): Energy of the beamline to scale reference calculations accordingly.

        Raises:
            Exception: Raised, when the scan_number is higher, then the total number of scans after averaging. --> Prints highest number possible.

        """
        
        # Set the references lists
        q_ref_array = []
        I_ref_array = []

        # Convert 2 teta to q for .int data
        def teta_to_q(teta,Energy):
            wavelenght = (6.62607015e-34*299792458*1e10)  / (Energy*1e3*1.602176634e-19)
            return ((4*np.pi*np.sin(np.deg2rad(teta)/2))/wavelenght)

        # Loop over all references
        for i in references:
            
            # Importing cif files with the xrayutilities module
            if references[i]['path'][-4:] == ".cif":
                
                # We import this module just if needed
                import xrayutilities as xu
                import os
                
                # make a crystal object from .cif
                xu_crystal = xu.materials.Crystal.fromCIF(os.path.join('data', references[i]['path']))
                
                # calculate the diffrection pattern for the set energy (croped to a cutoff Qmaxisnt~30 to reduce calculation time)
                pst = xu.simpack.PowderDiffraction(xu_crystal, en=energy*1e3, tt_cutoff=25)
                
                # Set the single reference lists 
                q_ref_j = []
                I_ref_j = []
                
                # Get data from the simulation
                for j in pst.data.values():
                    qpos_j = j["qpos"]
                    I_j = j["r"]
                    q_ref_j.append(qpos_j)
                    I_ref_j.append(I_j)
                    
                # Add peak broadening and exp decay.
                q_ref, I_ref = bars_to_gaussian(q_ref_j,I_ref_j, sigma=references[i]['broadening'])
                I_ref = exponential_radial_decay(I_ref, decay_rate=references[i]['decay_rate'])
                
                # Store in list
                q_ref_array.append(q_ref)
                I_ref_array.append(I_ref)
            

            # Alternativly one can import from .int files created by vista for a certain wavelenght.
            elif references[i]['path'][-4:] == ".int":
                
                # Load .int file
                q_ref, I_ref, unimp = np.loadtxt(references[i]['path'],skiprows=references[i]['skiprows']).T
                
                # Convert 2 teta to q and store in list
                q_ref_array.append(teta_to_q(q_ref,references[i]['energy']))
                I_ref_array.append(I_ref)
            
        # Add every plot (interactive matplotlib to uncheck rerferences)
        keys = [k for k, v in references.items()]
        fig, ax = plt.subplots(figsize=(self.figratio2D[0]*1.3, self.figratio2D[1]*1.3))
        
        # Add every reference to the plot
        lns = []
        for k in range(len(I_ref_array)):
            exec(f'l{k}, = ax.plot(q_ref_array[k],I_ref_array[k]/max(I_ref_array[k]), label=keys[k])')
            exec(f'lns.append(l{k})')
            
        # Give feedback number of the comparision scan is to high for the total scans of the data set (averaging shortens data amount)
        try:
            ax.plot(self.q,self.I[scan_number]/max(self.I[scan_number]))
        except:
            raise Exception('scan_number higher than the number of scans after averageing; Please select a smaller scan_number than: ' + str(len(self.I)))
       
        # Set Axis and title names
        ax.set_title('XRD phase check of ' + self.sampleName)    
        ax.set_xlabel("q [$\\rm{\AA^{-1}}$]")
        ax.set_ylabel("I [$\\rm{arb. un.}$]")
        
        # Crop data in X 
        if maxQ != None and minQ !=None:
            plt.xlim(minQ,maxQ)
        elif maxQ != None and minQ == None:
            plt.xlim(min(self.q),maxQ)
        elif minQ != None and maxQ == None:
            plt.xlim(minQ,max(self.q))
        elif maxQ == None and minQ == None:
            plt.xlim(min(self.q),max(self.q))
            
        # Crop data in Y 
        plt.ylim(bottom=0)
        
        # Adjust window layout
        plt.tight_layout()
        fig.subplots_adjust(right=0.7)

        # Link references checkboxes to plot color and label
        lines_by_label = {l.get_label(): l for l in lns}
        line_colors = [l.get_color() for l in lines_by_label.values()]

        # Make checkbuttons with all plotted lines with correct visibility
        rax = fig.add_axes([0.71, 0.5, 0.28, 0.07*len(I_ref_array)])
        check = CheckButtons(
            ax=rax,
            labels=lines_by_label.keys(),
            actives=[l.get_visible() for l in lines_by_label.values()],
            label_props={'color': line_colors},
            frame_props={'edgecolor': line_colors},
            check_props={'facecolor': line_colors},
            )
        
        # Update function for un-/checking boxes
        def callback(label):
            ln = lines_by_label[label]
            ln.set_visible(not ln.get_visible())
            ln.figure.canvas.draw_idle()

        # Update plot on click
        check.on_clicked(callback)
        plt.show()