import numpy as np
from time import time
import os, shutil
from scipy.stats import norm
from scipy import optimize
from scipy.signal import savgol_filter
from backbone.functions.whittaker_smooth import whittaker_smooth
    
#----------------------------------------------------------------------------------------#
#                  Ensures a output directory will be created on demand                  #
#----------------------------------------------------------------------------------------#

def ensure_dir(dir_path):
    """ Creates the probed directory if not existing
    Arguments
    ----------
    dir_path (str):   Path to directory.   
    """ 
    
    # Prob directory for existence
    if not os.path.exists(dir_path):
        # Create directory
        os.makedirs(dir_path)

#----------------------------------------------------------------------------------------#
#              Find closest related elemt to a given element in array                    #
#----------------------------------------------------------------------------------------#
def find_nearest(array, value):
    """Find element in array nearest to input value.  

    Args:
        array (list, numpy array): Array in which to search.
        value (float): Value to search for in array.

    Returns:
        float: Value in array which is nearest to the input value.
    """
    
    # If input is a list it will be converted to numpy array
    array = np.asarray(array)
    # Searches for the index the value closest to the wanted one in the array has.
    idx = (np.abs(array - value)).argmin()
    
    # Returns element closest to the value 
    return array[idx]

#----------------------------------------------------------------------------------------#
#                                   Get processing time                                  #
#----------------------------------------------------------------------------------------#
def processTime(t0):
    """ Defines the processing time from the starting time t0
    
    Arguments
    ----------
    t0 (date time): Start time.
    """
    
    # Prints to consol the time difference between now and t0.
    print("\n Time for processing")
    print("------------------------------------------------------------------")
    print("Time: " + str((time() - t0)/60))


#----------------------------------------------------------------------------------------#
#                     round to a multiple, to a value higher then 1                      #
#----------------------------------------------------------------------------------------#
def round_to_multiple(number,multiple):
    """ Rounds a number to the dividing factor given (multiple)
    
    Arguments
    ----------
    number (float): unrounded number
    multiple (float): number to round to
    
    Returns:
        rounded number (float)
    """
    return multiple * round(number/multiple)


#----------------------------------------------------------------------------------------#
#                                   get array from file                                   #
#----------------------------------------------------------------------------------------#
def getIfile(file):
    """ Loads q,I from a file with 1 header line.
    
    Arguments
    ----------
    file (str): path to array file
    
    Returns:
        intensity (numpy array)
    """
    q_file, j = np.loadtxt(file,skiprows=1).T
    return j


def averageArray(array,average: int, dimension: int =1):
    """ Average over a set number of scans inside a matrix. 
    Args:
        array (array): Input array (or matrix)
        average (int): Number of elements to average over
        dimension (int, optional): Dimension is 1 for vector like array and 2 for matrix like array. Defaults to 1.

    Returns:
        array: Averaged array (1 or 2 dimensional)
    """
    #Create new list for array
    new_array = []
    
    #Create a array for the average array
    if dimension == 1:
        array_averaged = 0
    if dimension == 2:
        array_averaged = np.zeros(len(array[0]))
    
    #Enumerate over all scans
    for i,j in enumerate(array):
        
        #Add up array of each scan 
        array_averaged = array_averaged + j
        
        #Store the averaged array when the averaging is reached and
        # normalize again to the average
        if (i+1) % average == 0 and i > 0:
            array_averaged = array_averaged/average  
            new_array.append(array_averaged)              
            if dimension == 1:
                array_averaged = 0
            if dimension == 2:
                array_averaged = np.zeros(len(array[0]))
            
    #Update array to the averaged data
    return new_array

def averageArrayError(array,average: int, dimension: int =1):
    """ Average over a set number of arrays that are related to the error of a different array inside a matrix. 

    Args:
        array (array): Input array (or matrix)
        average (int): Number of elements to average over
        dimension (int, optional): Dimension is 1 for vector like array and 2 for matrix like array. Defaults to 1.

    Returns:
        array: Averaged array (1 or 2 dimensional)
    """
    #Create new list for array
    new_array = []
    
    #Create a array for the average array
    if dimension == 1:
        array_averaged = 0
    if dimension == 2:
        array_averaged = np.zeros(len(array[0]))
    
    #Enumerate over all scans
    for i,j in enumerate(array):
        
        #Add up the squared error in each array of each scan 
        array_averaged = array_averaged + j**2
        
        #Store the averaged array when the averaging is reached and
        # normalize again to the root of the average minus 1
        if (i+1) % average == 0 and i > 0:
            array_averaged = np.sqrt(array_averaged/ average)  
            new_array.append(array_averaged)              
            if dimension == 1:
                array_averaged = 0
            if dimension == 2:
                array_averaged = np.zeros(len(array[0]))
            
    #Update array to the averaged data
    return new_array

def del_folder_content(dir_path):
    """ Delete folder content.

    Args:
        dir_path (str): Pat to the FOlder, which content should be deleted.
    """
    
    # Loop over the howl folder content
    for filename in os.listdir(dir_path):
        
        # Define file path for each file
        file_path = os.path.join(dir_path, filename)
        
        # Delete file
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
                
        # Error if file failes to be deleted for some reason
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))
            
def bars_to_gaussian(x_data, y_data, sigma=0.1):
    """
    Convert bar plot data (x_data, y_data) to a plot with Gaussian functions at bar positions.

    Parameters:
    - x_data: X-axis data (list or array)
    - y_data: Y-axis data (list or array)
    - sigma: Standard deviation of the Gaussian functions (default is 0.1)

    Returns:
    - x: q array
    - y: Intensity(q) array
    
    """
    
    # Create the varible for the sum of all gauss curves 
    sum = 0
    
    # Create the common X-Axis for each curve
    x = np.linspace(0,max(x_data)+1,10000)
    
    # Iterate through bar positions and heights to plot Gaussian functions
    for position, height in zip(x_data, y_data):
        # Create the gauss curve
        y = height * norm.pdf(x, position, sigma)
        # Add the gauss curve to the sum
        sum += y
    # Normalize the sum of all gauss curves to 1    
    sum = sum / max(sum)
    return x,sum

def exponential_radial_decay(original_array, decay_rate):
    """
    Apply exponential radial decay to dampen an array.

    Parameters:
    - original_array (list or numpy array): The input array to be dampened 
    - decay_rate (float): The decay rate for exponential decay 

    Returns:
    - Dampened array (numpy array)
    """
    
    # Get the full leghth 
    array_length = len(original_array)
    
    # Create the linear X-Axis
    distances = np.arange(array_length)
    
    # Subtract the dampening function from the original data set
    dampened_array =  original_array * np.exp(-decay_rate*1e-3 * distances) #original_array * (1 - decay_rate*1e-1* distances**(1/4))
    return dampened_array


def overwrite_file(file_path, data_array1, data_array2, header_lines: int=27):
    """
    Overwrite a file with two data arrays, excluding the header lines.

    Parameters:
    file_path (str): The path of the file to be overwritten.
    data_array1 (list or numpy array): The first data array to be written to the file.
    data_array2 (list or numpy array): The second data array to be written to the file.
    header_lines (int, optional): Header lines from the original file to be kept.
    """
    # Combine data arrays
    combined_data = np.column_stack((data_array1, data_array2))

    # Read header lines
    with open(file_path, 'r') as file:
        original_lines = file.readlines()

    # Write data to the file, excluding the header lines
    with open(file_path, 'w') as file:
        file.writelines(original_lines[:header_lines])
        for row in combined_data:
            file.write(' '.join(map(str, row)) + '\n')
            
# Fit function for the FT noise (sine function)
def fit_sin(tt, yy):
    """Fit sin to the input time sequence, and return fitting parameters "amp", "omega", "phase", "offset", "freq", "period" and "fitfunc"

    Args:
        tt (list, array): X-Axis of data the sin fit is applied to.
        yy (list, array): Y-Axis of data the sin fit is applied to.

    Returns:
        c (float): offset in y.
        p (float): phase coefficient.
    """
    
    # Working with numpy arrays to handel lists
    tt = np.array(tt)
    yy = np.array(yy)
    
    
    ff = np.fft.fftfreq(len(tt), (tt[1]-tt[0]))   # assume uniform spacing
    Fyy = abs(np.fft.fft(yy))
    guess_freq = abs(ff[np.argmax(Fyy[1:])+1])   # excluding the zero frequency "peak", which is related to offset
    guess_amp = np.std(yy) * 2.**0.5
    guess_offset = np.mean(yy)
    guess = np.array([guess_amp, 2.*np.pi*guess_freq, 0., guess_offset])

    # Classic sine function with phase p, y-offset c  and amplitude A
    def sinfunc(t, A, w, p, c):  
        return A * np.sin(w*t + p) + c
    
    # Using scipy optimize for fitting
    popt, pcov = optimize.curve_fit(sinfunc, tt, yy, p0=guess)
    
    # Extract the varibles from the fit output
    A, w, p, c = popt

    # Reduce the phase term to be within 2*pi
    multiple_of_2pi = np.floor(p / (2 * np.pi))
    p = p - multiple_of_2pi * (2 * np.pi)
    
    return c, p



def fit_modulated_wave(t, y):
    """
    Fit a modulated wave and return the carrier wave's phase shift.

    Parameters:
    - t: Time array
    - y: Observed signal

    Returns:
    - Phase shift of the carrier wave
    """
    
    def modulated_wave(t, A, x0, f_mod, f_carrier, phase_mod, phase_carrier):
        """
        Function representing a modulated wave.

        Parameters:
        - t: Time array
        - A: Amplitude of the modulated wave
        - f_mod: Modulation frequency
        - f_carrier: Carrier frequency
        - phase_mod: Modulation phase
        - phase_carrier: Carrier phase

        Returns:
        - Modulated wave signal
        """
        modulation =  A * np.sin(f_mod * t + phase_mod)
        carrier = np.sin(f_carrier * t + phase_carrier)
        return modulation * carrier + x0
    
    # Initial guess for parameters (amplitude, modulation frequency, carrier frequency, mod phase, carrier phase)
    initial_guess = [1.0, 0.0, 2.0, 10.0, 0.0, 0.0]

    # Perform the fit
    params, covariance = optimize.curve_fit(modulated_wave, t, y, p0=initial_guess)

    # Extract the carrier phase from the fitted parameters
    phase_carrier = params[-1]
    
    import matplotlib.pyplot as plt 
    plt.plot(t,modulated_wave(t,params[0],params[1],params[2],params[3],params[4],params[5]))
    plt.show()
    
    return phase_carrier


def shorten_matrix_elements(matrix,limits: list):
    """ Takes a matrix and cuts every array in the matrix according to a given lower and upper limit.

    Args:
        matrix (list of lists or list of numpy array): Matrix which array elements should be shortend.
        limits (list): list with [lower limit, upper limit] of the array elements.
    """
    
    for i,j in enumerate(matrix):
        matrix[i] = j[limits[0]:limits[1]]

def Filter(y,Filter,strength,order: int=2):
        """ Use a Savatzky-Golay or Whittaker-Eilers filter for the XRD/TS data.
        
        Arguments
        ----------
        Filter:     Type of Filter to use (lFilter or savgol)
        strength:   Strength with what the filter is applied 
        order:      Order of polinominal magnitude 
        """
        # Applying a filter if set
        if Filter == "savgol":
            y = savgol_filter(y,strength,order)
        elif Filter == "whittaker":
            y = whittaker_smooth(y,strength,d=order)
            
        return y
